# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------


# This file is a python class representation of the ASN.1 schema defined
# in manifest.asn. It is used to generate a JSON representation of the manifest
# from raw data.
#
# It differs from manifest_definition.py as it's not used by pyasn1, it's not used
# to encode the manifest data nor is it automatically generated.
#
# The process looks like:
#   (Input data) -> (Python object representation) -> (JSON) -> (PyASN1 object) -> ([DER] encoding)
#                             [ this file ]
import time, collections
from manifesttool import utils
from manifesttool.errorhandler import InvalidObject

class BaseObject(object):

    # Basic check to see if all required fields are set.
    # Does NOT check if the set value is of right type accoridng to ASN.1 definition
    def verify(self, required_fields):
        for f in required_fields:
            if self.__dict__.get(f) is None:
                raise InvalidObject("Value for field %r not set, but it's required" % f)

    # Convinience method to just convert object to python dict
    def to_dict(self):
        return utils.todict(self)

class SignedResource(BaseObject):
    def __init__(self, resource, signature):
        self.resource = resource
        self.signature = signature


class CertificateReference(object):
    def __init__(self, fingerprint, uri):
        self.fingerprint = fingerprint
        self.uri = uri

class SignatureBlock(object):
    def __init__(self, signature, certificates):
        self.signature = signature
        self.certificates = certificates

class ResourceSignature(object):
    def __init__(self, hash, signatures):
        self.hash = hash
        self.signatures = signatures



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
from __future__ import division
import codecs, hashlib, sys, json, os, base64, binascii, logging, ecdsa, uuid
from collections import namedtuple

# from pyasn1 import debug
# debug.setLogger(debug.Debug('all'))

from cryptography.hazmat.primitives import ciphers as cryptoCiphers
from cryptography.hazmat.primitives import hashes as cryptoHashes
from cryptography.hazmat.primitives.ciphers.aead import AESCCM
from cryptography.hazmat import backends as cryptoBackends
from cryptography import x509

from manifesttool import utils, codec, errorhandler, defaults
from manifesttool.v1.manifest_schema import SignedResource, Resource, ResourceSignature, ResourceReference, \
                                         Manifest, CertificateReference, PayloadDescription, ResourceAlias, \
                                         SignatureBlock, MacBlock
from manifesttool.v1 import manifest_definition
from manifesttool import keytable_pb2


# Import different encoders
from pyasn1.codec.der import encoder as der_encoder
from pyasn1.codec.native import decoder as native_decoder
from pyasn1.error import PyAsn1Error
from pyasn1.type.univ import OctetString, Sequence
from pyasn1.type import namedtype

from builtins import bytes

nonceSize = 16

LOG = logging.getLogger(__name__)

def manifestGet(manifest, path, shortpath = None):
    if shortpath:
        fieldValue = manifestGet(manifest, path = shortpath)
        if fieldValue:
            return fieldValue
    pathList = path.split('.')
    subManifest = manifest
    while subManifest and pathList:
        subManifest = subManifest.get(pathList[0])
        pathList = pathList[1:]
    if subManifest and not isinstance(subManifest, (str, bytes, int, bool, dict, list)):
        subManifest = str(subManifest)
    return subManifest

class cryptoMode:
    _mode = namedtuple('CryptoMode', 'name,symSize,hashSize,payloadEncrypt')
    MODES = {
        0: _mode('invalid', None, None, None),
        1: _mode('aes-128-ctr-ecc-secp256r1-sha256', 128/8, 256/8, True),
        2: _mode('none-ecc-secp256r1-sha256', 128/8, 256/8, False),
        3: _mode('none-none-sha256', 128/8, 256/8, False),
        4: _mode('none-psk-aes-128-ccm-sha256', 128/8, 256/8, False),
        5: _mode('psk-aes-128-ccm-sha256', 128/8, 256/8, True)
    }
    @staticmethod
    def name2enum(name):
        for k,v in cryptoMode.MODES.items():
            if v.name == name:
                return k
        else:
            return None
    def __init__(self, mode):
        m = self.MODES.get(mode)
        self.symSize = m.symSize
        self.hashSize = m.hashSize
        self.payloadEncrypt = m.payloadEncrypt
        self.name = m.name

def ARM_UC_mmCryptoModeShouldEncrypt(mode):
    return cryptoMode(mode).payloadEncrypt


def encryptClearText(cleartext, key, iv):
    backend = cryptoBackends.default_backend()
    iv_counter = int.from_bytes(bytes(iv), byteorder='big')
    # Create the cipher object
    cipher = cryptoCiphers.Cipher(cryptoCiphers.algorithms.AES(key), cryptoCiphers.modes.CTR(iv), backend=backend)
    encryptor = cipher.encryptor()
    # Encrypt the plain text
    ciphertext = encryptor.update(bytes(cleartext)) + encryptor.finalize()
    return ciphertext

def encryptKeyAES(options, manifestInput, encryptionInfo, payload):
    # Store the PSK ID
    pskId = manifestGet(manifestInput, 'resource.resource.manifest.payload.encryptionInfo.id.key', 'encryptionKeyId')
    if not pskId:
        LOG.critical('The aes-psk encryption mode requires a "encryptionKeyId" entry in the input JSON file')
        sys.exit(1)
    encryptionInfo["id"] = { "key" : codecs.decode(pskId, 'hex')}
    secret = options.payload_secret

    # Read the secret
    secret_data = utils.read_file(secret) if os.path.isfile(secret) else base64.b64decode(secret)
    if not secret_data:
        LOG.critical('The aes-psk encryption mode requires either a base64-encoded secret or a valid file to be passed via the "-s" option.')
        sys.exit(1)

    # generate an IV
    IV = os.urandom(nonceSize)
    # generate a new key
    key = os.urandom(nonceSize)

    cipherpayload = encryptClearText(payload, key, IV)

    # NOTE: In general, ECB is a poor choice, since it fails quickly to known-plaintext attacks. In this case, however,
    # ECB should be appropriate, since the plaintext is precisely one AES block, and it is comprised of high entropy
    # random data; therefore, the plaintext is never known.

    # Create the cipher object
    cipher = cryptoCiphers.Cipher(cryptoCiphers.algorithms.AES(bytes(secret_data)),
                                  cryptoCiphers.modes.ECB(), backend=backend) # nosec ignore bandit B404
    keyEncryptor = cipher.encryptor()
    # Encrypt the plain text
    cipherKey = keyEncryptor.update(bytes(key)) + encryptor.finalize()
    return ciphertext

    return IV, cipherKey, cipherpayload

def get_payload_description(options, manifestInput):
    crypto_mode = manifestGet(manifestInput, 'resource.resource.manifest.encryptionMode.enum', 'encryptionMode')
    crypto_mode = crypto_mode if crypto_mode in cryptoMode.MODES else cryptoMode.name2enum(crypto_mode)
    if not crypto_mode:
        crypto_mode = get_crypto_mode(options, manifestInput)
    crypto_name = cryptoMode.MODES.get(crypto_mode).name

    payload_file = manifestGet(manifestInput, 'resource.resource.manifest.payload.reference.file', 'payloadFile')
    if hasattr(options, 'payload') and options.payload:
        payload_file = options.payload.name
    # payload URI defaults to the payload file path if no payload URI is supplied.
    payload_uri = payload_file
    payload_uri = manifestGet(manifestInput, 'resource.resource.manifest.payload.reference.uri', 'payloadUri')
    if hasattr(options, 'uri') and options.uri:
        payload_uri = options.uri

    dependencies = manifestGet(manifestInput, 'resource.resource.manifest.dependencies')
    if not any((payload_uri, payload_file)) and not dependencies:
        LOG.critical('No payload was specified and no dependencies were provided.')
        sys.exit(1)
        # fwFile/fwUri is optional, so if not provided we just return empty
        return None

    payload_hash = manifestGet(manifestInput, 'resource.resource.manifest.payload.reference.hash', 'payloadHash')
    payload_size = manifestGet(manifestInput, 'resource.resource.manifest.payload.reference.size', 'payloadSize')
    if payload_hash:
        LOG.debug('Found hash in input, skipping payload load. Hash: {}'.format(payload_hash))
        payload_hash = binascii.a2b_hex(payload_hash)
    else:
        if not payload_file:
            LOG.critical('Either a payload file or a payload hash must be specified.')
            sys.exit(1)
        payload_filePath = payload_file
        # If file path is not absolute, then make it relative to input file
        if os.path.isabs(payload_file) or not options.input_file.isatty():
            payload_filePath = os.path.join(os.path.dirname(options.input_file.name), payload_file)
        # Read payload input, record length and hash it
        payload_size = utils.size_file(options, payload_filePath)
        payload_hash = utils.hash_file(options, payload_filePath)
        LOG.debug('payload of {} bytes loaded. Hash: {}'.format(payload_size, binascii.b2a_hex(payload_hash)))

    # Ensure the cryptoMode is valid
    if not crypto_mode in cryptoMode.MODES:
        valid_modes = ", ".join((('%s (%d)' % (v.name, k) for k, v in cryptoMode.MODES.items())))
        LOG.critical('Could not find specified cryptoMode (%d) in list of valid encryption modes. '
            'Please use on of the following: %r' % (crypto_mode, valid_modes))
        sys.exit(1)

    # Get encryption options for the provided mode
    should_encrypt = ARM_UC_mmCryptoModeShouldEncrypt(crypto_mode)
    if hasattr(options, 'encrypt_payload') and options.encrypt_payload and not should_encrypt:
        LOG.critical('--encrypt-payload specified, but cryptoMode({cryptoMode}) does not support encryption.'.format(**manifestInput))
        sys.exit(1)

    encryptionInfo = None
    if should_encrypt:
        LOG.debug('Crypto mode {} ({}) requires encryption. Will ensure ciphers are valid and loaded...'\
                .format(crypto_mode, crypto_name))
        if not options.encrypt_payload:
            LOG.critical('Specified crypto mode ({cryptoMode}) requires encryption, '
                        'but --encrypt-payload not specified.'.format(**manifestInput))
            sys.exit(1)
        cipherFile = manifestGet(manifestInput, 'resource.resource.manifest.payload.encryptionInfo.file', 'encryptedPayloadFile')
        if not cipherFile:
            LOG.critical('"resource.resource.manifest.payload.encryptionInfo.file" must be specified in the JSON input'
                        'file when --encrypt-payload is specified on the command-line.')
            sys.exit(1)
        encryptionInfo = {}

        cipherModes = {
                'aes-psk' : encryptKeyAES
        }

        if not options.encrypt_payload in cipherModes:
            LOG.critical('Specified encryption mode "{mode}" is not supported'.format(mode=options.encrypt_payload))
            sys.exit(1)

        init_vector, cipherKey, cipherpayload = cipherModes.get(options.encrypt_payload)(options, manifestInput, encryptionInfo, content)

        with open(cipherFile,'wb') as f:
            f.write(cipherpayload)

        encryptionInfo["key"] = { "cipherKey": cipherKey }
        encryptionInfo["initVector"] = init_vector
        LOG.debug('payload ({} bytes) encrypted. Cipher key: {}, cipher payload ouptut to : {}'.format(len(content), cipherKey, cipherFile))

    else:
        LOG.debug('Will not encrypt payload as crypto mode {} ({}) does not require it'.format(crypto_mode, crypto_name))

    return PayloadDescription(
        **{
            "storageIdentifier": manifestGet(manifestInput,'resource.resource.manifest.payload.storageIdentifier', 'storageIdentifier') or "default",
            "reference": ResourceReference(
                hash = payload_hash,
                uri = payload_uri,
                size = payload_size
            ),
            "encryptionInfo": encryptionInfo
        }
    )

def get_manifest_aliases(options, manifestInput):
    aliases = []
    for alias in manifestGet(manifestInput,'resource.resource.manifest.aliases') or []:
        # Ensure all required options are defined
        if not all(k in alias for k in ('file', 'uri')):
            LOG.critical('Could not create aliases, as all required keys for alias are not defined ("uri" and "file")')
            sys.exit(1)

        # Read alias file to calculate hash
        fPath = alias['file']
        if not os.path.isabs(fPath):
            fPath = os.path.join(os.path.dirname(options.input_file.name), fPath)
        content = utils.read_file(fPath)
        sha_hash = utils.sha_hash(content)
        LOG.debug('Creating ResourceAlias for file {} with SHA reference {}'.format(fPath, sha_hash))

        aliases.append(ResourceAlias(
            hash = sha_hash,
            uri = alias['uri']
        ))
    return aliases

def get_manifest_dependencies(options, manifestInput):
    dependencies = []
    for link in manifestGet(manifestInput, 'resource.resource.manifest.dependencies') or []:
        if not any(k in link for k in ('uri', 'file')):
            LOG.critical('Manifest link requires either a "uri" or a "file"'
                        'key - or both. Could only find %r' % link.keys())
            sys.exit(1)
        LOG.debug('Adding manifest link reference (URI: {}, File: {})'.format(link.get('uri', None), link.get('file', None)))

        # If file isn't provided, we attempt to download the file from provided URI
        # If the user provides the link as a filepath, just read the file
        if 'file' in link:
            linkPath = link['file']
            if not os.path.isabs(linkPath):
                linkPath = os.path.join(os.path.dirname(options.input_file.name), linkPath)
            content = utils.read_file(linkPath)
        else:
            content = utils.download_file(link['uri'])

        # Calculate the hash and length.
        manifest_hash = utils.sha_hash(content)
        manifest_length = len(content)
        LOG.debug('Linked manifest of {} bytes loaded. Hash: {}'.format(manifest_length, manifest_hash))

        link_uri = link['uri'] if 'uri' in link else link['file']
        dependencies.append({
            'hash': manifest_hash,
            'uri': link_uri,
            'size': manifest_length
        })
    return dependencies

def get_crypto_mode(options, manifestInput):
    crypto_mode = manifestGet(manifestInput, 'resource.resource.manifest.encryptionMode.enum', 'encryptionMode')
    crypto_mode = crypto_mode if crypto_mode in cryptoMode.MODES else cryptoMode.name2enum(crypto_mode)
    if not crypto_mode:
        isEncrypted = hasattr(options, 'encrypt_payload') and options.encrypt_payload
        isMAC = hasattr(options, 'mac') and options.mac

        if isEncrypted and isMAC:
            cryptname = 'psk-aes-128-ccm-sha256'
        if not isEncrypted and isMAC:
            cryptname = 'none-psk-aes-128-ccm-sha256'
        if not isEncrypted and not isMAC:
            cryptname = 'none-ecc-secp256r1-sha256'
        if isEncrypted and not isMAC:
            cryptname = 'aes-128-ctr-ecc-secp256r1-sha256'
        crypto_mode = cryptoMode.name2enum(cryptname)

    return crypto_mode

def get_manifest(options, manifestInput):
    vendor_info = manifestGet(manifestInput, 'resource.resource.manifest.vendorInfo', 'vendorInfo') or b""
    vendor_id = manifestGet(manifestInput, 'resource.resource.manifest.vendorId', 'vendorId')
    device_id = manifestGet(manifestInput, 'resource.resource.manifest.deviceId', 'deviceId')
    class_id = manifestGet(manifestInput, 'resource.resource.manifest.classId', 'classId')

    if not device_id and not (vendor_id and class_id):
        LOG.critical('Input file must contain either Device ID, or both Vendor and Class IDs, or all three\n'
            'deviceId:"{}"\nvendorId:"{}"\nclassId:"{}"'.format(device_id, vendor_id, class_id))
        sys.exit(1)

    validTo = manifestGet(manifestInput, 'resource.resource.manifest.applyPeriod.validTo', 'applyPeriod.validTo')
    validFrom = manifestGet(manifestInput, 'resource.resource.manifest.applyPeriod.validFrom', 'applyPeriod.validFrom')
    applyPeriod = None
    if validTo or validFrom:
        applyPeriod = {
            'validFrom' : validFrom,
            'validTo'   : validTo,
        }

    crypto_mode = { 'enum' : get_crypto_mode(options, manifestInput) }
    nonce = manifestGet(manifestInput, 'resource.resource.manifest.nonce')
    if nonce:
        nonce = binascii.a2b_hex(nonce)
    else:
        nonce = os.urandom(nonceSize)
    if not manifestGet(manifestInput, 'resource.resource.manifest.applyImmediately', 'applyImmediately'):
        LOG.warning('applyImmediately is currently ignored by the update client; manifests are always applied immediately.')
    return Manifest(
        manifestVersion = manifestGet(manifestInput,'resource.resource.manifest.version'),
        vendorInfo = vendor_info,
        vendorId = uuid.UUID(vendor_id).bytes if vendor_id else b'',
        classId = uuid.UUID(class_id).bytes if class_id else b'',
        deviceId = uuid.UUID(device_id).bytes if device_id else b'',
        nonce = nonce,
        applyImmediately = manifestGet(manifestInput, 'resource.resource.manifest.applyImmediately', 'applyImmediately'),
        applyPeriod = applyPeriod,
        encryptionMode = crypto_mode,
        description = manifestGet(manifestInput, 'resource.resource.manifest.description', 'description'),
        aliases = get_manifest_aliases(options, manifestInput),
        payload = get_payload_description(options, manifestInput),
        dependencies = get_manifest_dependencies(options, manifestInput),
        timestamp = manifestGet(manifestInput, 'resource.resource.manifest.timestamp')
    )

def getDevicePSKData(mode, psk, nonce, digest, payload_key):
    crypto_mode = cryptoMode(mode)
    LOG.debug('Creating PSK data: ')
    LOG.debug('    mode: {}'.format(crypto_mode.name))
    LOG.debug('    iv  ({0} bytes): {1}'.format(len(nonce), binascii.b2a_hex(nonce)))
    LOG.debug('    md  ({0} bytes): {1}'.format(len(digest), binascii.b2a_hex(digest)))

    data = {'hash':str(digest)}
    if payload_key:
        data['cek'] = payload_key
    asnData = native_decoder.decode(data, asn1Spec=manifest_definition.KeyTableEntry())
    # print(asnData)

    plainText = der_encoder.encode(asnData)
    LOG.debug('PSK plaintext ({0} bytes): {1}'.format(len(plainText), binascii.b2a_hex(plainText)))

    pskSig = {
        'none-psk-aes-128-ccm-sha256': lambda key, iv, d: AESCCM(key[:128//8]).encrypt(iv, d, b''),
        'psk-aes-128-ccm-sha256': lambda key, iv, d: AESCCM(key[:128//8]).encrypt(iv, d, b'')
    }.get(crypto_mode.name, lambda a,b,d: None)(psk, nonce, plainText)
    LOG.debug('PSK data ({0} bytes): {1}'.format(len(pskSig), binascii.b2a_hex(pskSig)))
    return pskSig

def symmetricArgsOkay(options):
    okay = hasattr(options, 'private_key') and options.private_key
    # okay = okay and hasattr(options, 'psk_table_encoding') and options.psk_table_encoding
    if hasattr(options, 'psk_table_encoding') and options.psk_table_encoding:
        okay = okay and True
    okay = okay and hasattr(options, 'psk_table') and options.psk_table
    return okay

def isPsk(mode):
    crypto_mode = cryptoMode(mode)
    return {
        'aes-128-ctr-ecc-secp256r1-sha256' : False,
        'none-ecc-secp256r1-sha256':False,
        'none-none-sha256': False,
        'none-psk-aes-128-ccm-sha256': True,
        'psk-aes-128-ccm-sha256': True
    }.get(crypto_mode.name)

def get_symmetric_signature(options, manifestInput, enc_data):
    input_hash = manifestGet(manifestInput,'signature.hash') or b''

    # There should always be a signing key on create.
    if not hasattr(options,'private_key') or not options.private_key:
        if 'psk-master-key' in manifestInput:
            try:
                options.private_key = open(manifestInput['psk-master-key'],'rb')
            except:
                LOG.critical('No PSK master key specified and default key ({}) cannot be opened'.format(manifestInput['private-key']))
                sys.exit(1)
        else:
            LOG.critical('Resource is not signed and no PSK master key is provided.')
            sys.exit(1)

    if not symmetricArgsOkay(options):
        LOG.critical('--mac requires:')
        LOG.critical('    --private-key: The master key for generating device PSKs')
        #                                                                                 80 chars ->|
        #             ================================================================================
        LOG.critical('    --psk-table: the output file for PSKs. This argument is optional/ignored')
        LOG.critical('                 when inline encoding is used.')
        LOG.critical('    --psk-table-encoding: the encoding to use for the PSK output file.')
        LOG.critical('    Either:')
        LOG.critical('        --device-urn: The device URN for the target device (Endpoint Client Name)')
        LOG.critical('    OR:')
        LOG.critical('        --filter-id: The filter identifier used to gather device URNs')
        sys.exit(1)

    # Get SHA-256 hash of content and sign it using private key
    sha_content = utils.sha_hash(enc_data)
    # If a hash is provided in the input json, then the encoded content must match the provided hash
    if input_hash and sha_content != binascii.a2b_hex(input_hash):
        LOG.critical('Manifest hash provided in input file does not match hashed output')
        LOG.critical('Expected: {0}'.format(input_hash))
        LOG.critical('Actual:   {0}'.format(binascii.b2a_hex(sha_content)))
        sys.exit(1)

    # Load a default URN out of the settings file.
    devices = manifestGet(manifestInput,'deviceURNs') or []
    LOG.info('Loaded device URNs from {!r}'.format(defaults.config))
    for dev in devices:
        LOG.info('    {!r}'.format(dev))

    # Optionally load a URN out of the command-line arguments
    if hasattr(options, 'device_urn') and options.device_urn:
        cmdDevs = [options.device_urn]
        LOG.info('Loaded device URNs from input arguments')
        for dev in cmdDevs:
            LOG.info('    {!r}'.format(dev))
        devices.extend(cmdDevs)
    if hasattr(options, 'filter_id') and options.filter_id:
        LOG.critical('Device Filters not supported yet.')
        sys.exit(1)

    # Use only unique devices
    devices = list(set(devices))

    crypto_mode = get_crypto_mode(options, manifestInput)

    payload_key = b''
    if hasattr(options, 'payload_key') and options.payload_key:
        LOG.debug('Converting payload key ({0}) to binary'.format(options.payload_key))
        payload_key = bytes(binascii.a2b_hex(options.payload_key))
    LOG.debug('Payload key length: {0}'.format(len(payload_key)))
    master_key = options.private_key.read()
    vendor_id = manifestGet(manifestInput, 'resource.resource.manifest.vendorId', 'vendorId')
    class_id = manifestGet(manifestInput, 'resource.resource.manifest.classId', 'classId')
    iv = os.urandom(13)
    deviceSymmetricInfos = {}
    # For each device
    maxIndexSize = 0
    maxRecordSize = 0
    LOG.info('Creating per-device validation codes...')
    for device in devices:
        LOG.info('    {!r}'.format(device))

        hkdf = utils.getDevicePSK_HKDF(cryptoMode(crypto_mode).name, master_key, uuid.UUID(vendor_id).bytes, uuid.UUID(class_id).bytes, b'Authentication')
        psk = hkdf.derive(bytes(device, 'utf-8'))
        maxIndexSize = max(maxIndexSize, len(device))
        # Now encrypt the hash with the selected AE algorithm.
        pskCipherData = getDevicePSKData(crypto_mode, psk, iv, sha_content, payload_key)
        recordData = der_encoder.encode(OctetString(hexValue=binascii.b2a_hex(pskCipherData).decode("ascii")))
        maxRecordSize = max(maxRecordSize, len(recordData))
        deviceSymmetricInfos[device] = recordData
    # print (deviceSymmetricInfos)
    def proto_encode(x):
        keytable = keytable_pb2.KeyTable()

        for k, d in x.items():
            entry = keytable.entries.add()
            entry.urn = k
            entry.opaque = d
        return keytable.SerializeToString()
    # Save the symmetric info file
    encodedSymmertricInfos = {
        'json' : lambda x : json.JSONEncoder(default=binascii.b2a_base64).encode(x),
        'cbor' : lambda x : None,
        'protobuf': proto_encode,
        'text' : lambda x : '\n'.join([','.join([binascii.b2a_hex(y) for y in (k,d)]) for k, d in x.items()]) + '\n'
        # 'inline' : lambda x : None
    }.get(options.psk_table_encoding)(deviceSymmetricInfos)
    options.psk_table.write(encodedSymmertricInfos)

    #=========================
    # PSK ID is the subject key identifier (hash) of the master key.
    # The URI of the "certificate" is the location at which to find the key table or key query.
    # PSK is known only to the signer and the device
    # PSK Signature is AE(PSK, hash), but it is not included, since it must be distributed in the key table.
    shaMaster = cryptoHashes.Hash(cryptoHashes.SHA256(), cryptoBackends.default_backend())
    shaMaster.update(master_key)
    subjectKeyIdentifier = shaMaster.finalize()

    mb = MacBlock(   pskID = subjectKeyIdentifier,
                keyTableIV = iv,
               keyTableRef = 'thismessage://1',
           keyTableVersion = 0,
        keyTableRecordSize = maxRecordSize,
         keyTableIndexSize = maxIndexSize
    )
    macs=[mb]
    rs = ResourceSignature(
            hash = sha_content,
            signatures = [],
            macs = macs)
    return rs


def get_signature(options, manifestInput, enc_data):
    signatures = manifestGet(manifestInput,'signature.signatures') or []
    input_hash = manifestGet(manifestInput,'signature.hash') or b''

    # There should always be a signing key on create.
    if not hasattr(options,'private_key') or not options.private_key:
        if 'private-key' in manifestInput:
            try:
                options.private_key = open(manifestInput['private-key'],'r')
            except:
                LOG.critical('No private key specified and default key ({}) cannot be opened'.format(manifestInput['private-key']))
                sys.exit(1)
        else:
            LOG.critical('Resource is not signed and no signing key is provided.')
            sys.exit(1)

    # Get SHA-256 hash of content and sign it using private key
    sha_content = utils.sha_hash(enc_data)
    if len(signatures):
        # If a signature is provided in the input json, then the encoded content must match the provided hash
        # Signature validation is not performed, since this would require certificate acquisition, which may not be
        # possible
        if sha_content != binascii.a2b_hex(input_hash):
            LOG.critical('Manifest hash provided in input file does not match hashed output')
            LOG.critical('Expected: {0}'.format(input_hash))
            LOG.critical('Actual:   {0}'.format(binascii.b2a_hex(sha_content)))
            sys.exit(1)
        # TODO: perform best-effort signature validation

    if hasattr(options, 'private_key') and options.private_key:
        sk = ecdsa.SigningKey.from_pem(options.private_key.read())
        sig = sk.sign_digest(sha_content, sigencode=ecdsa.util.sigencode_der)

        certificates = []

        # pick a signature block with no signature in it.
        inputCerts = manifestGet(manifestInput, 'certificates') or []

        # If no certificate was provided in the manifest input or in options,
        if len(inputCerts) == 0:
            # then load the default certificate
            inputCerts = manifestInput.get('default-certificates', [])

        # If there is still no certificate,
        if len(inputCerts) == 0:
            # Search through all signature blocks for one that contains certificates but no signature
            for idx, sb in enumerate(signatures):
                 if not 'signature' in sb:
                     inputCerts = sb.get('certificates', [])
                     # This signature will be appended later so we must trim it.
                     del signatures[idx]
                     break

        for idx, cert in enumerate(inputCerts):
            if not any(k in cert for k in ('file', 'uri')):
                LOG.critical('Could not find "file" or "uri" property for certificate')
                sys.exit(1)

            # If 'file', we just use the content in local file
            if 'file' in cert:
                fPath = cert['file']
                if not os.path.isabs(fPath):
                    fPath = os.path.join(os.path.dirname(options.input_file.name), cert['file'])
                content = utils.read_file(fPath)

            # Else we download the file contents
            else:
                content = utils.download_file(cert['uri'])
            # Figure our which extension the certificate has
            contentPath = cert['file'] if 'file' in cert else cert['uri']
            ext = contentPath.rsplit('.', 1)[1]

            # Read the certificate file, and get DER encoded data
            if ext == 'pem':
                lines = content.replace(" ",'').split()
                content = binascii.a2b_base64(''.join(lines[1:-1]))

            # Verify the certificate hash algorithm
            # Extract subjectPublicKeyInfo field from X.509 certificate (see RFC3280)
            # fingerprint = utils.sha_hash(content)
            cPath = cert['file'] if 'file' in cert else cert['uri']
            certObj = None
            try:
                certObj = x509.load_der_x509_certificate(content, cryptoBackends.default_backend())
            except ValueError as e:
                LOG.critical("X.509 Certificate Error in ({file}): {error}".format(error=e, file=cPath))
                sys.exit(1)

            if not certObj:
                LOG.critical("({file}) is not a valid certificate".format(file=cPath))
                sys.exit(1)
            if not isinstance(certObj.signature_hash_algorithm, cryptoHashes.SHA256):
                LOG.critical("In ({file}): Only SHA256 certificates are supported by the Mbed Cloud Update client at this time.".format(file=cPath))
                sys.exit(1)
            fingerprint = certObj.fingerprint(cryptoHashes.SHA256())

            LOG.debug('Creating certificate reference ({}) from {} with fingerprint {}'.format(idx, contentPath, fingerprint))
            uri = ''
            if 'uri' in cert:
                uri = cert['uri']
            certificates.append(CertificateReference(
                fingerprint = fingerprint,
                uri = uri
            ))

        LOG.debug('Signed hash ({}) of encoded content ({} bytes) with resulting signature {}'.format(
            sha_content, len(enc_data), sig))
        signatures.append(SignatureBlock(signature = sig, certificates = certificates))
    return ResourceSignature(
            hash = sha_content,
            signatures = signatures
        )

def create_signed_resource(options):
    # Read options from manifest input file/
    # if (options.input_file.isatty()):
    #     LOG.info("Reading data from from active TTY... Terminate input with ^D")
    manifestInput = {
        'applyImmediately' : True
    }

    try:
        if os.path.exists(defaults.config):
            with open(defaults.config) as f:
                manifestInput.update(json.load(f))
        if not options.input_file.isatty():
            content = options.input_file.read()
            if content and len(content) >= 2: #The minimum size of a JSON file is 2: '{}'
                manifestInput.update(json.loads(content))
    except ValueError as e:
        LOG.critical("JSON Decode Error: {}".format(e))
        sys.exit(1)

    # Create the Resource structure first, and encode it using specified encoding scheme
    # This encoded data will then be used for signing
    try:
        resource = Resource(
            resource = get_manifest(options, manifestInput),
            resourceType  = Resource.TYPE_MANIFEST
        )
    except errorhandler.InvalidObject as err:
        LOG.critical('Unable to create resource object: {0}'.format(err))
        sys.exit(1)

    # Convert the Python object to a Python dictionary, and then encode to
    # specified encoding format
    resource_encoded = utils.encode(resource.to_dict(), options, manifest_definition.Resource())

    signature = None
    crypto_mode = get_crypto_mode(options, manifestInput)
    if isPsk(crypto_mode):
        signature = get_symmetric_signature(options, manifestInput, enc_data = resource_encoded)
    else:
        signature = get_signature(options, manifestInput, enc_data = resource_encoded)
    try:
        sresource = SignedResource(
            resource = resource,
            signature = signature
        )
        return sresource.to_dict()
    except errorhandler.InvalidObject as err:
        LOG.critical('Unable to create signed resource: {0}'.format(err))
        sys.exit(1)

def toCfile(varname, data, colwidth=120, indentsize=4):
    s  = '#include <stdint.h>\n'
    s += 'uint8_t {}[] = {{\n'.format(varname)
    xl = ['{0:#04x}'.format(b) for b in bytes(data)]
    # w = (ax + b*(x-1))
    # w = ax + bx - b
    # (w + b) / (a + b)
    rowcount = int((colwidth - indentsize + len(', ') - len(',')) / len(', 0x00'))
    while len(xl):
        rownum = min(rowcount,len(xl))
        row = '    ' + ', '.join(xl[:rownum])
        if len(xl) > rownum:
            row += ','
        xl = xl[rownum:]
        row += '\n'
        s += row;
    s += '};\n'
    return codecs.encode(s, 'utf8')

def write_result(data, options):
    if hasattr(options,'hex') and options.hex:
        LOG.debug('Converting binary data into hex encoded string')
        data = codecs.encode(data, 'hex')
    if hasattr(options,'c_file') and options.c_file:
        LOG.debug('Converting binary data into a C file.')
        data = toCfile('manifest', data)

    # Write result to output file or stdout buffer.
    options.output_file.write(data)

    # Append newline if outputting to TTY
    if hasattr(options, 'output_file') and options.output_file.isatty():
        options.output_file.write(b'\n')

def main(options):
    LOG.debug('Creating new manifest from input file and options')

    # Parse input files and options. Generate hydrated and hierachial manifest JSON.
    manifest = create_signed_resource(options)
    LOG.debug('Manifest python object successfully created from ASN.1 definition and input')

    # Encode data if requested
    output = utils.encode(manifest, options, manifest_definition.SignedResource())
    LOG.debug('Manifest successfully encoded into desired format ({}). Size: {} bytes.'.format(options.encoding, len(output)))

    if hasattr(options, 'mac') and options.mac and options.psk_table_encoding == 'inline':
        output += options.psk_table

    # And we're done
    write_result(output, options)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

# This file has been generated using asn1ate (v <unknown>) from './tools/manifest-wrapper.asn'
# Last Modified on 2017-01-06 14:48:42
from pyasn1.type import univ, char, namedtype, namedval, tag, constraint, useful


class Bytes(univ.OctetString):
    pass


class Uri(char.UTF8String):
    pass


class CertificateReference(univ.Sequence):
    pass


CertificateReference.componentType = namedtype.NamedTypes(
    namedtype.NamedType('fingerprint', Bytes()),
    namedtype.NamedType('uri', Uri())
)


class SignatureBlock(univ.Sequence):
    pass


SignatureBlock.componentType = namedtype.NamedTypes(
    namedtype.NamedType('signature', univ.OctetString()),
    namedtype.NamedType('certificates', univ.SequenceOf(componentType=CertificateReference()))
)

class MacBlock(univ.Sequence):
    pass


MacBlock.componentType = namedtype.NamedTypes(
    namedtype.NamedType('pskID', univ.OctetString()),
    namedtype.NamedType('keyTableVersion', univ.Integer()),
    namedtype.OptionalNamedType('keyTableIV', univ.OctetString()),
    namedtype.OptionalNamedType('keyTableRef', char.UTF8String()),
    namedtype.NamedType('keyTableIndexSize', univ.Integer()),
    namedtype.NamedType('keyTableRecordSize', univ.Integer())
)

class ResourceSignature(univ.Sequence):
    pass


ResourceSignature.componentType = namedtype.NamedTypes(
    namedtype.NamedType('hash', univ.OctetString()),
    namedtype.NamedType('signatures', univ.SequenceOf(componentType=SignatureBlock())),
    namedtype.OptionalNamedType('macs', univ.SequenceOf(componentType=MacBlock()))
)


class SignedResource(univ.Sequence):
    pass


SignedResource.componentType = namedtype.NamedTypes(
    namedtype.NamedType('resource', univ.Any()),
    namedtype.NamedType('signature', ResourceSignature())
)


class Payload(univ.OctetString):
    pass


class UUID(univ.OctetString):
    pass



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
from __future__ import print_function, division
from builtins import input, bytes, chr
import os, sys, json, logging, uuid, re
import binascii
from manifesttool import defaults, cert, utils, templates
from manifesttool.argparser import MainArgumentParser
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.backends import default_backend

LOG = logging.getLogger(__name__)

def writeUpdateDefaults(options):
    if os.path.isfile(defaults.updateResources) and not (hasattr(options, 'force') and options.force):
        LOG.warning('{} already exists, not overwriting. Use --force to write anyway.'.format(defaults.updateResources))
        return
    with open(defaults.updateResources, 'w') as defaultResources:
        # Format the device UUIDs
        vendorId = ', '.join(['0x%x' % x for x in bytes(uuid.UUID(options.vendor_id).bytes)])
        classId = ', '.join(['0x%x' % x for x in bytes(uuid.UUID(options.class_id).bytes)])

        certFp = ''
        cert = ''
        str_ski = ''
        # Read the certificate
        if not hasattr(options, 'psk') or not options.psk:
            options.certificate.seek(0)
            cstr = options.certificate.read()
            try:
                # Load the certificate.
                certObj = x509.load_der_x509_certificate(
                    cstr,
                    default_backend()
                )
            except ValueError as e:
                raise ValueError('Error loading {}: {}'.format(options.certificate.name, e.message))
            # Calculate the certificate fingerprint
            c_hash = certObj.fingerprint(hashes.SHA256())
            # Format the certificate fingerprint
            certFp = ', '.join(['0x%x' % x for x in bytes(c_hash[:16])]) + ',\n    ' + ', '.join(['0x%x' % x for x in bytes(c_hash[16:])])

            # Calculate the subjectKeyIdentifier
            c_ski = certObj.extensions.get_extension_for_oid(x509.oid.ExtensionOID.SUBJECT_KEY_IDENTIFIER).value.digest
            str_ski = ', '.join(['0x%x' % x for x in bytes(c_ski[:16])]) + ',\n    ' + ', '.join(['0x%x' % x for x in bytes(c_ski[16:])])

            # Format the certificate
            CByteArray = [ '0x%x'% x for x in bytes(cstr)]
            CLineList = [ ', '.join(CByteArray[i:i+16]) for i in range(0,len(CByteArray),16)]
            cert = ',\n    '.join(CLineList)

        defaultResources.write(templates.UpdateDefaultResources.format(
                vendorId = vendorId,
                classId = classId,
                certFp = certFp,
                ski = str_ski,
                cert = cert,
                psk = ', '.join(['0x%x'%x for x in bytes(options.psk)]),
                pskId = ', '.join(['0x%x'%x for x in bytes(options.psk_id)])
        ))
        LOG.info('Wrote default resource values to {}'.format(defaults.updateResources))

def mkCert(options):
        cmd = ['cert', 'create', '-o', defaults.certificate, '-K', defaults.certificateKey]

        country = ''
        state = ''
        locality = ''
        organization = ''
        commonName = ''
        if hasattr(options, 'vendor_domain') and options.vendor_domain:
            commonName = options.vendor_domain
        validity = defaults.certificateDuration

        if not options.quiet:
            print('A certificate has not been provided to init, and no certificate is provided in {cert}'.format(
                cert=defaults.certificate))
            print('Init will now guide you through the creation of a certificate.')
            print()
            print('This process will create a self-signed certificate, which is not suitable for production use.')
            print()
            print('In the terminology used by certificates, the "subject" means the holder of the private key that matches a certificate.')
            country = input('In which country is the subject located? ').strip()
            state = input('In which state or province is the subject located? ').strip()
            locality = input('In which city or region is the subject located? ').strip()
            organization = input('What is the name of the subject organization? ').strip()
            commonName = ''
            if hasattr(options, 'vendor_domain') and options.vendor_domain:
                commonName = input('What is the common name of the subject organization? [{}]'.format(options.vendor_domain)).strip() or options.vendor_domain
            else:
                commonName = input('What is the common name of the subject organization? ')
            validity = input('How long (in days) should the certificate be valid? [{}]'.format(defaults.certificateDuration)).strip() or defaults.certificateDuration

        try:
            os.makedirs(defaults.certificatePath)
        except os.error:
            # It is okay if the directory already exists. If something else went wrong, we'll find out when the
            # create occurs
            pass

        cmd = ['cert', 'create', '-o', defaults.certificate, '-K', defaults.certificateKey,
            '-V', str(validity)]
        if country:
            cmd += ['-C', country]
        if state:
            cmd += ['-S', state]
        if locality:
            cmd += ['-L', locality]
        if organization:
            cmd += ['-O', organization]
        if commonName:
            cmd += ['-U', commonName]
        cert_opts = MainArgumentParser().parse_args(cmd).options
        rc = cert.main(cert_opts)
        if rc:
            sys.exit(1)
        options.certificate = open(defaults.certificate, 'rb')
        LOG.info('Certificate written to {}'.format(defaults.certificate))
        options.private_key = open(defaults.certificateKey, 'rb')
        LOG.info('Private key written to {}'.format(defaults.certificateKey))

def checkURN(deviceURN):
    URNsplit = deviceURN.split(':')
    if len(URNsplit) < 3 or URNsplit[0] != 'urn':
        raise ValueError('PSK Identity does not appear to be a valid URN: {!r}'.format(deviceURN))
    if URNsplit[1] not in ['dev', 'uuid', 'imei', 'esn', 'meid', 'imei-msisdn', 'imei-imsi']:
        raise ValueError('{!r} is not a recommended URN class for PSK identities')
    if URNsplit[1] == 'dev' and URNsplit[2] != 'ops':
        raise ValueError('ops-type URNs are recommended dev-class URNs for PSK identities')

def findDevCertFiles(textPattern, directorySearchPaths, searchExtension):
    IdentityFiles = []
    AllSourceFiles = []
    for path in directorySearchPaths:
        if os.path.isdir(path):
            for file in os.listdir(path):
                if file.endswith(searchExtension):
                    AllSourceFiles.append(os.path.join(path,file))
    for file in AllSourceFiles:
        with open(file, 'rt') as fd:
            for line in fd:
                if textPattern in line:
                    IdentityFiles.append(file)
    if len(IdentityFiles) > 1:
        raise ValueError('Multiple Endpoint Name definitions found!')
        sys.exit(1)
    else:
        if len(IdentityFiles) == 1:
            return IdentityFiles[0]
        else:
            return ''


def main(options):

    settings = {}
    # Check if a settings file exists
    if os.path.isfile(defaults.config):
        # load default settings
        with open(defaults.config,'r') as f:
            settings = json.load(f)
    # Populate a default list of PSKIdentities
    PSKIdentities = settings.get('deviceURNs', [])

    if hasattr(options, 'vendor_domain') and options.vendor_domain:
        domainParts = options.vendor_domain.split('.')
        minPart = min(domainParts)
        if len(domainParts) < 2 or len(minPart) < 1:
            LOG.critical('"{0}" is not a valid domain name.'.format(options.vendor_domain))
            return 1
        options.vendor_id = str(uuid.uuid5(uuid.NAMESPACE_DNS, options.vendor_domain))
    vendorId = uuid.UUID(options.vendor_id)

    if hasattr(options, 'model_name') and options.model_name:
        options.class_id = str(uuid.uuid5(vendorId, options.model_name))
    classId = uuid.UUID(options.class_id)

    if hasattr(options, 'psk') and options.psk:
        # Attempt to read the device endpoint name out of mbed_cloud_dev_credentials.c
        deviceURN = None
        try:
            dev_credential_dirs = ['.','source']
            dev_cred_file_name = findDevCertFiles('MBED_CLOUD_DEV_BOOTSTRAP_ENDPOINT_NAME', dev_credential_dirs, '.c')
            with open(dev_cred_file_name, 'rt') as fd:
                URNLine = None
                for line in fd:
                    if 'MBED_CLOUD_DEV_BOOTSTRAP_ENDPOINT_NAME' in line:
                        URNLine = line
                        break
                if URNLine:
                    m = re.search('const\s+char\s+MBED_CLOUD_DEV_BOOTSTRAP_ENDPOINT_NAME\\[\\]\s*=\s*"([^"]*)";', URNLine)
                    if m:
                        deviceURN = m.group(1)
        except IOError:
            # Handle missing file
            pass
        # A device URN is mandatory for psk init. The URN can come from either
        # the list in .manifest_tool.json or from the command-line. It is
        # debatable whether init should be used if the .manifest_tool.json is
        # already populated, since the purpose of manifest-tool init is to fill
        # in .manifest_tool.json.
        if not deviceURN and len(PSKIdentities) == 0:
            LOG.critical('An endpoint name must be populated in mbed_cloud_dev_credentials.c')
            return 1
        if deviceURN == '0':
            LOG.critical('An endpoint name in mbed_cloud_dev_credentials.c must be initialized (not \'0\') and must be different for each device.')
            return 1
        if deviceURN in PSKIdentities:
            LOG.warning('Device URN %r already exists in .manifest-tool.json. Is it unique?', deviceURN)

        try:
            checkURN(deviceURN)
        except ValueError as e:
            LOG.warning('%s', str(e))

        # Append the device URN extracted from the credentials file to the current list
        PSKIdentities.append(deviceURN)
        # Install the device URN extracted from the credentials file into the options object for
        # use in generating template files
        options.device_urn = deviceURN
        masterKey = None
        masterKeyRequired = not hasattr(options, 'master_key') or not options.master_key
        if masterKeyRequired:
            try:
                # Open and store to master_key since this matches what argparse is doing
                options.master_key = open(defaults.pskMasterKey, 'rb')
                masterKeyRequired = False
                if options.force:
                    LOG.warning("Using existing master key in %r. Not generating a new master key.",
                        defaults.pskMasterKey)
            except:
                masterKeyRequired = True

        if masterKeyRequired:
            LOG.info('Generating a new 256-bit master key')
            masterKey = os.urandom(256//8)
            masterKeyName = defaults.pskMasterKey
            try:
                os.makedirs(defaults.certificatePath)
            except os.error:
                # It is okay if the directory already exists. If something else went wrong, we'll find out when the
                # create occurs
                pass
            # Open and store to master_key since this matches what argparse is doing
            options.master_key = open(masterKeyName,'wb')
            LOG.info('Storing master key to %r', options.master_key.name)
            options.master_key.write(masterKey)
            options.master_key.flush()
            options.master_key.seek(0)
        else:
            LOG.info('Reading master key out of {}'.format(options.master_key.name))
            masterKey = options.master_key.read()

        # Generate the device PSK
        shaMaster = hashes.Hash(hashes.SHA256(), default_backend())
        shaMaster.update(masterKey)
        options.psk_id = shaMaster.finalize()
        psk_hkdf = utils.getDevicePSK_HKDF('none-psk-aes-128-ccm-sha256', masterKey, vendorId.bytes, classId.bytes, b'Authentication')
        options.psk = psk_hkdf.derive(bytes(options.device_urn, 'utf-8'))

    else:
        cert_required = True
        options.psk_id = bytes('', 'utf-8')
        options.psk = bytes('', 'utf-8')
        options.device_urn = ''
        if options.certificate:
            cert_required = False
        elif hasattr(options,'force') and options.force:
            cert_required = True
        else:
            try:
                options.certificate = open(defaults.certificate,'rb')
                options.private_key = open(defaults.certificateKey, 'rb')
                cert_required = False
            except:
                cert_required = True

        if cert_required:
            mkCert(options)
    # Write the settings

    settings = {
        'classId' : str(classId),
        'vendorId' : str(vendorId),
        'vendorDomain' : options.vendor_domain,
        'modelName' : options.model_name,
        'deviceURNs' : list(set(PSKIdentities))
    }

    if hasattr(options, 'psk') and options.psk:
        settings['psk-master-key'] = options.master_key.name
    else:
        settings['private-key'] = options.private_key.name
        settings['default-certificates'] = [ {'file':options.certificate.name}]

    with open(defaults.config, 'w') as f:
        f.write(json.dumps(settings, sort_keys=True, indent=4))
        LOG.info('Default settings written to {}'.format(defaults.config))

    try:
        writeUpdateDefaults(options)
    except ValueError as e:
        LOG.critical('Error setting defaults: {}'.format(e.message))
        return 1

    cloud_settings = {}
    if hasattr(options, 'server_address') and options.server_address:
        cloud_settings['host'] = options.server_address
    if hasattr(options, 'api_key') and options.api_key:
        cloud_settings['api_key'] = options.api_key

    if cloud_settings:
        with open(defaults.cloud_config, 'w') as f:
            f.write(json.dumps(cloud_settings, sort_keys=True, indent=4))
            LOG.info('Cloud settings written to {}'.format(defaults.cloud_config))

    return 0



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import os
certificatePath = '.update-certificates'
certificate = os.path.join(certificatePath,'default.der')
certificateKey = os.path.join(certificatePath,'default.key.pem')
certificateDuration = 90
pskMasterKey = os.path.join(certificatePath,'default.master.psk')
config = '.manifest_tool.json'
cloud_config = '.mbed_cloud_config.json'
updateResources = 'update_default_resources.c'



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: keytable.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='keytable.proto',
  package='pskManifest',
  syntax='proto3',
  serialized_pb=_b('\n\x0ekeytable.proto\x12\x0bpskManifest\"*\n\x0b\x44\x65viceEntry\x12\x0b\n\x03urn\x18\x01 \x01(\t\x12\x0e\n\x06opaque\x18\x02 \x01(\x0c\"5\n\x08KeyTable\x12)\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x18.pskManifest.DeviceEntryb\x06proto3')
)




_DEVICEENTRY = _descriptor.Descriptor(
  name='DeviceEntry',
  full_name='pskManifest.DeviceEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='urn', full_name='pskManifest.DeviceEntry.urn', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='opaque', full_name='pskManifest.DeviceEntry.opaque', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=31,
  serialized_end=73,
)


_KEYTABLE = _descriptor.Descriptor(
  name='KeyTable',
  full_name='pskManifest.KeyTable',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='entries', full_name='pskManifest.KeyTable.entries', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=75,
  serialized_end=128,
)

_KEYTABLE.fields_by_name['entries'].message_type = _DEVICEENTRY
DESCRIPTOR.message_types_by_name['DeviceEntry'] = _DEVICEENTRY
DESCRIPTOR.message_types_by_name['KeyTable'] = _KEYTABLE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DeviceEntry = _reflection.GeneratedProtocolMessageType('DeviceEntry', (_message.Message,), dict(
  DESCRIPTOR = _DEVICEENTRY,
  __module__ = 'keytable_pb2'
  # @@protoc_insertion_point(class_scope:pskManifest.DeviceEntry)
  ))
_sym_db.RegisterMessage(DeviceEntry)

KeyTable = _reflection.GeneratedProtocolMessageType('KeyTable', (_message.Message,), dict(
  DESCRIPTOR = _KEYTABLE,
  __module__ = 'keytable_pb2'
  # @@protoc_insertion_point(class_scope:pskManifest.KeyTable)
  ))
_sym_db.RegisterMessage(KeyTable)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

UpdateDefaultResources = '''
//----------------------------------------------------------------------------
//   The confidential and proprietary information contained in this file may
//   only be used by a person authorised under and to the extent permitted
//   by a subsisting licensing agreement from ARM Limited or its affiliates.
//
//          (C) COPYRIGHT 2013-2017 ARM Limited or its affiliates.
//              ALL RIGHTS RESERVED
//
//   This entire notice must be reproduced on all copies of this file
//   and copies of this file may only be made by a person if such person is
//   permitted to do so under the terms of a subsisting license agreement
//   from ARM Limited or its affiliates.
//----------------------------------------------------------------------------

#ifdef MBED_CLOUD_CLIENT_USER_CONFIG_FILE
#include MBED_CLOUD_CLIENT_USER_CONFIG_FILE
#endif

#include <stdint.h>

#ifdef MBED_CLOUD_DEV_UPDATE_ID
const uint8_t arm_uc_vendor_id[] = {{
    {vendorId}
}};
const uint16_t arm_uc_vendor_id_size = sizeof(arm_uc_vendor_id);

const uint8_t arm_uc_class_id[]  = {{
    {classId}
}};
const uint16_t arm_uc_class_id_size = sizeof(arm_uc_class_id);
#endif

#ifdef MBED_CLOUD_DEV_UPDATE_CERT
const uint8_t arm_uc_default_fingerprint[] =  {{
    {certFp}
}};
const uint16_t arm_uc_default_fingerprint_size =
    sizeof(arm_uc_default_fingerprint);

const uint8_t arm_uc_default_subject_key_identifier[] =  {{
    {ski}
}};
const uint16_t arm_uc_default_subject_key_identifier_size =
    sizeof(arm_uc_default_subject_key_identifier);

const uint8_t arm_uc_default_certificate[] = {{
    {cert}
}};
const uint16_t arm_uc_default_certificate_size = sizeof(arm_uc_default_certificate);
#endif


#ifdef MBED_CLOUD_DEV_UPDATE_PSK
const uint8_t arm_uc_default_psk[] = {{
    {psk}
}};
const uint8_t arm_uc_default_psk_size = sizeof(arm_uc_default_psk);
const uint16_t arm_uc_default_psk_bits = sizeof(arm_uc_default_psk)*8;

const uint8_t arm_uc_default_psk_id[] = {{
    {pskId}
}};
const uint8_t arm_uc_default_psk_id_size = sizeof(arm_uc_default_psk_id);
#endif
'''



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import logging, os, binascii, json
from manifesttool import verify
from manifesttool.v1 import verify_signed_resource as definition
from manifesttool import codec, utils, defaults
from manifesttool import signature_schema as schema
from manifesttool import create
# All supported decoders
from pyasn1.codec.der import decoder as der_decoder

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.exceptions import InvalidSignature
from cryptography import x509

LOG = logging.getLogger(__name__)

def sig_from_dict(d):
    certificates = []
    for cr in d['certificates']:
        certificates.append(schema.CertificateReference(
            fingerprint = binascii.a2b_hex(cr['fingerprint']),
                    uri = cr.get('uri','')
        ))
    return schema.SignatureBlock(
           signature = binascii.a2b_hex(d['signature']),
        certificates = certificates)

def sign(options):
    # Load defaults
    defaultConfig = {}
    if os.path.exists(defaults.config):
        with open(defaults.config) as f:
            defaultConfig = json.load(f)

    # Load the existing manifest
    content = None
    if hasattr(options, 'manifest') and options.manifest:
        options.manifest.seek(0)
        content = options.manifest.read()
    else:
        content = utils.download_file(options.url)

    # Extract the signed resource.
    signed_resource_data = {
        "der": lambda d: codec.bin2obj(d, definition.SignedResource(), der_decoder),
    }[options.encoding](content)
    if not signed_resource_data:
        return 1
    LOG.debug('Decoded SignedResource from {} encoded binary'.format(options.encoding))

    # Sign the resource
    # Calculate the hash of the content of the existing manifest
    c_hash = utils.sha_hash(signed_resource_data['resource'])
    if not 'hash' in signed_resource_data['signature']:
        LOG.critical('Manifest does not contain a hash')
        return None
    # Extract the hash
    e_hash = binascii.a2b_hex(signed_resource_data['signature']['hash'])
    # Compare calculated and extracted hashes
    if c_hash != e_hash:
        LOG.critical('Hash mismatch\nExpected: {}\nActual:   {}'.format(binascii.b2a_hex(e_hash),binascii.b2a_hex(c_hash)))
        return None
    LOG.debug('Manifest hash: {}'.format(binascii.b2a_hex(c_hash)))

    # Load a certificate and private key
    # Private key must be in .manifest_tool.json, or provided on the command-line

    # 1. Check the command-line
    if not hasattr(options,'private_key') or not options.private_key:
        # 2. Check the default config
        if 'private-key' in defaultConfig:
            try:
                # NOTE: binary is not specified since the key is usually PEM encoded.
                options.private_key = open(defaultConfig['private-key'],'r')
            except:
                LOG.critical('No private key specified and default key ({}) cannot be opened'.format(defaultConfig['private-key']))
                return 1
    # 3. Fail if the private key is not found
    if not hasattr(options, 'private_key') or not options.private_key:
        LOG.critical('No private key specified and default key ({}) cannot be opened'.format(defaultConfig['private-key']))
        return 1
    if not hasattr(options, 'password'):
        options.password = None

    # Load the private key
    privkey = load_pem_private_key(options.private_key.read(), password=options.password, backend=default_backend())

    # Make sure that this is an ECDSA key!
    if not isinstance(privkey, ec.EllipticCurvePrivateKey):
        LOG.critical('Private key was not an ECC private key')
        return 1
    LOG.debug('Loaded private key')
    LOG.info('Signing manifest...')
    # Create signature
    sig = privkey.sign(signed_resource_data['resource'], ec.ECDSA(hashes.SHA256()))
    LOG.debug('Signature: {}'.format(binascii.b2a_hex(sig)))

    # destroy the privkey object
    privkey = None

    # Certificate must be in .manifest_tool.json, or provided on the command-line
    # Load the certificate
    if not hasattr(options,'certificate') or not options.certificate:
        if 'default-certificates' in defaultConfig:
            options.certificate = open(defaultConfig['default-certificates'][0]['file'],'rb')

    if not hasattr(options,'certificate') or not options.certificate:
        LOG.critical('No certificate specified and default certificate ({}) cannot be opened'.format(defaultConfig['default-certificates'][0]['file']))
        return 1

    # Load the certificate object from the DER file
    certObj = None
    try:
        certObj = x509.load_der_x509_certificate(options.certificate.read(), default_backend())
    except ValueError as e:
        LOG.critical("X.509 Certificate Error in ({file}): {error}".format(error=e, file=options.certificate.name))
        return(1)

    if not certObj:
        LOG.critical("({file}) is not a valid certificate".format(file=cPath))
        return(1)

    # Make sure that the certificate is signed with SHA256
    if not isinstance(certObj.signature_hash_algorithm, hashes.SHA256):
        LOG.critical("In ({file}): Only SHA256 certificates are supported by the update client at this time.".format(file=cPath))
        return(1)

    LOG.info('Verifying signature with supplied certificate...')

    # Verify the signature with the provided certificate to ensure that the update target will be able to do so
    try:
        pubkey = certObj.public_key()
        pubkey.verify(sig, signed_resource_data['resource'], ec.ECDSA(hashes.SHA256()))
    except InvalidSignature as e:
        LOG.critical('New signature failed to verify with supplied certificate ({})'.format(options.certificate.name))
        return 1

    # Store the fingerprint of the certificate
    fingerprint = certObj.fingerprint(hashes.SHA256())

    certificates = []
    # for idx in range(len())
    #     certObj = None
    #     try:
    #         certObj = x509.load_der_x509_certificate(options.certificate.read(), default_backend())
    #     except ValueError as e:
    #         LOG.critical("X.509 Certificate Error in ({file}): {error}".format(error=e, file=options.certificate.name))
    #         return(1)
    #
    #     if not certObj:
    #         LOG.critical("({file}) is not a valid certificate".format(file=cPath))
    #         return(1)
    #     if not isinstance(certObj.signature_hash_algorithm, hashes.SHA256):
    #         LOG.critical("In ({file}): Only SHA256 certificates are supported by the update client at this time.".format(file=cPath))
    #         return(1)
    #     LOG.debug('Creating certificate reference ({}) from {} with fingerprint {}'.format(idx, options.certificate.name, fingerprint))
    LOG.debug('Creating certificate reference from {} with fingerprint {}'.format(options.certificate.name, fingerprint))
    uri = ''
    # TODO: Insert URI for delegation of trust
    cr = schema.CertificateReference(
        fingerprint = fingerprint,
        uri = uri
    )
    # Append the certificate reference to the current list
    # NOTE: Currently, only one certificate reference will exist in the certificates list, but with delegation of trust
    #       there will be more certificate references
    certificates.append(cr)
    LOG.debug('Signed hash ({}) of encoded content ({} bytes) with resulting signature {}'.format(
        binascii.b2a_hex(c_hash), len(content), binascii.b2a_hex(sig)))

    signatures = []
    for s in signed_resource_data['signature']['signatures']:
        signatures.append(sig_from_dict(s))

    # encode the signature block
    signatures.append(schema.SignatureBlock(signature = sig, certificates = certificates))

    # encode the resource signature
    resource_signature = schema.ResourceSignature(
            hash = c_hash,
            signatures = signatures
        )

    # encode the signed resource
    signed_resource = schema.SignedResource(
        resource = signed_resource_data['resource'],
        signature = resource_signature
    )

    # Convert the signed resource into a python dictionary
    manifest_dict = signed_resource.to_dict()
    # Encode the Python dictionary as a DER stream
    output = utils.encode(manifest_dict, options, definition.SignedResource())

    # Write the result to the output_file
    if hasattr(options, 'output_file') and options.output_file:
        # Write result to output file or stdout buffer.
        options.output_file.write(output)

        # Append newline if outputting to TTY
        if options.output_file.isatty():
            options.output_file.write(b'\n')
        return 0
    return 1

def main(options):
    if hasattr(options, 'manifest') and options.manifest:
        options.input_file = options.manifest
    if not hasattr(options, 'input_file') or not options.input_file:
        return 1
    rc = verify.main(options)
    if rc:
        return rc
    LOG.debug('Adding new signature to manifest')
    return sign(options)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import codecs, uuid, binascii
from pyasn1.codec.native.decoder import decode as native_decode
from pyasn1.codec.native.encoder import encode as native_encode

def get_component_type(schema):
    """Due to quirks in the pyasn1 library, for some schema types the componentType is not
    properly set, so we use this convinience method to access the underlying data if required"""
    return schema.componentType if (len(schema.componentType) > 0) else schema._componentType

def obj2asn(value, schema):
    """
    Converts python object into the pyasn1 object following the schema provided. The (potentially nested)
    object (commenly converted from JSON) will of course have to match the format of the ASN.1 defintion.

    Given the following schema:

        MyStruct SEQUENCE ::= {
            id INTEGER,
            name UTF8String
        }

    Example:

        >>> import schema_definition
        >>> from pyasn1.codec.der import encoder
        >>> value = JSON.loads('{ "id": 1, "name": "David" }')
        >>>
        >>> # pyasn1 schema can be generated from ASN.1 definition using asn2py script, here
        >>> # exported as schema_definition.py, which we can import as a python module.
        >>> asn_obj = obj2asn(value, schema_definition.MyStruct())
        >>>
        >>> # We can now encode this data to whatever format we want
        >>> der_bytes = encoder.encode(asn_obj)
    """
    return native_decode(value, schema)

def asn2obj(asn, numericEnum=False):
    """
    Decodes ASN objects into Python object structures (array or dictionary). See
    opposing function `obj2asn` for the encoding logic.

    This is commonly called for when you have a pyasn1 object (e.g. as a result
    of running as pyasn1 decode from raw bytes) and you want to output into JSON
    or similar.

    Example, given the same schema as defined in `obj2asn`:

        >>> from pyasn1.codec.der import decoder
        >>> import schema_definition
        >>> import json
        >>> raw_bytes = b"..."
        >>> asn_obj = decoder.decode(raw_bytes, schema_definition.MyStruct())
        >>> print json.dumps(asn2obj(asn_obj), indent=4)
        {
            "id": 1,
            "name": "David"
        }
    """

    if asn.__class__.__name__ == 'Any':
        return asn.asOctets()

    # If type_id is not set, we have a primitive type
    if not asn.typeId:
        # If the type has base type which is UTF8String, we use the string
        # representation of these bytes - and not the raw representation as
        # value. Might be a better way of checking if it's a string class, but
        # this works fine.
        if 'UTF8String' in [b.__name__ for b in asn.__class__.__bases__] or \
            asn.__class__.__name__ == 'UTF8String':
            if isinstance(asn._value, bytes):
                return asn._value.decode(asn._encoding)
            return str(asn._value)

        # Translate the UUID type into URN format (RFC-4122)
        if asn.__class__.__name__ == 'UUID':
            if len(asn._value) != 16:
                return binascii.b2a_hex(asn.asOctets())
            else:
                return str(uuid.UUID(bytes=asn._value))

        # The type OBJECT IDENTIFIER becomes a tuple type, so we
        # decode it into a "." delimited string
        if isinstance(asn._value, tuple):
            return '.'.join(map(str, asn._value))

        # The ASN type Boolean becomes 1/0, but we would like to translate
        # to Python booleans.
        if asn.__class__.__name__ == 'Boolean':
            return bool(asn._value)

        if asn.__class__.__name__ == 'Enumerated':
            if numericEnum:
                return int(asn._value)
            else:
                return asn.getNamedValues().getName(asn)
        # If we have bytes, we encode into hex representation. Hacky, but we
        # need something JSON serializable.
        if isinstance(asn._value, bytes):
            return codecs.encode(asn._value, 'hex').decode(asn._encoding)

        return asn._value

    # If type_id is 1 or 2 we have a Sequence/SetOf type (i.e. an array)
    elif asn.typeId <= 2:
        return [asn2obj(e, numericEnum) for e in asn._componentValues]

    # If not primitive nor array, we have a Sequence-like type and we handle
    # this as a dictionary. We iterate through each key where the value is
    # set (i.e. ignoring empty optional fields) and recursivly get the decoded
    # value block.
    j = {}
    if hasattr(asn, '_componentValues'):
        for idx, v in enumerate(asn._componentValues):
            if v is None or (hasattr(v,'__len__') and len(v) is 0): continue
            j[asn.getNameByPosition(idx)] = asn2obj(v,numericEnum)
    return j

def asn2obj_native(asn):
    """
    Decodes ASN objects into Python object structures (array or dictionary). See
    opposing function `obj2asn` for the encoding logic.

    This is commonly called for when you have a pyasn1 object (e.g. as a result
    of running as pyasn1 decode from raw bytes) and you want to output into JSON
    or similar.

    Example, given the same schema as defined in `obj2asn`:

        >>> from pyasn1.codec.der import decoder
        >>> import schema_definition
        >>> import json
        >>> raw_bytes = b"..."
        >>> asn_obj = decoder.decode(raw_bytes, schema_definition.MyStruct())
        >>> print json.dumps(asn2obj(asn_obj), indent=4)
        {
            "id": 1,
            "name": "David"
        }
    """
    return native_encode(asn)

def bin2obj(value, schema, decoder, numericEnum = False):
    decoded_asn = decoder.decode(value, schema)[0]
    return asn2obj(decoded_asn, numericEnum)

def bin2obj_native(value, schema, decoder):
    decoded_asn = decoder.decode(value, schema)[0]
    return asn2obj_native(decoded_asn)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
from __future__ import print_function
import logging, sys, os
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from six import text_type as text
import datetime
import binascii
from manifesttool import defaults
from manifesttool import utils

LOG = logging.getLogger(__name__)

def create(options):
    key = ec.generate_private_key(ec.SECP256R1(), default_backend())
    options.key_output.write(key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption()
    ))
    # Various details about who we are. For a self-signed certificate the
    # subject and issuer are always the same.
    subjectList = []
    try:
        if options.country:
            subjectList += [x509.NameAttribute(x509.oid.NameOID.COUNTRY_NAME, text(options.country))]
        if options.state:
            subjectList += [x509.NameAttribute(x509.oid.NameOID.STATE_OR_PROVINCE_NAME, text(options.state))]
        if options.locality:
            subjectList += [x509.NameAttribute(x509.oid.NameOID.LOCALITY_NAME, text(options.locality))]
        if options.organization:
            subjectList += [x509.NameAttribute(x509.oid.NameOID.ORGANIZATION_NAME, text(options.organization))]
        if options.common_name:
            subjectList += [x509.NameAttribute(x509.oid.NameOID.COMMON_NAME, text(options.common_name))]
        if len(subjectList) == 0:
            subjectList += [x509.NameAttribute(x509.oid.NameOID.COMMON_NAME, text('localhost'))]
        subject = issuer = x509.Name(subjectList)
    except ValueError as e:
        LOG.critical('Error creating certificate: {}'.format(e.message))
        fname = options.output_file.name
        options.output_file.close()
        os.remove(fname)
        return 1

    subjectKey = key.public_key().public_bytes(serialization.Encoding.DER, serialization.PublicFormat.SubjectPublicKeyInfo)
    subjectKeyIdentifier = utils.sha_hash(subjectKey)[:160//8] # Use RFC7093, Method 1

    try:
        cert = x509.CertificateBuilder().subject_name(
                subject
            ).issuer_name(
                issuer
            ).public_key(
                key.public_key()
            ).serial_number(
                x509.random_serial_number()
            ).not_valid_before(
                datetime.datetime.utcnow()
            ).not_valid_after(
                datetime.datetime.utcnow() + datetime.timedelta(days=options.valid_time)
            ).add_extension(
                x509.KeyUsage(
                     digital_signature = True,
                    content_commitment = False,
                      key_encipherment = False,
                     data_encipherment = False,
                         key_agreement = False,
                         key_cert_sign = False,
                              crl_sign = False,
                         encipher_only = False,
                         decipher_only = False),
                critical=False
            ).add_extension(
                x509.SubjectAlternativeName([x509.DNSName(u"localhost")]),
                critical=False,
            ).add_extension(
                x509.ExtendedKeyUsage([x509.oid.ExtendedKeyUsageOID.CODE_SIGNING]),
                critical=False,
            ).add_extension(
                x509.SubjectKeyIdentifier(subjectKeyIdentifier),
                critical=False,
            # Sign our certificate with our private key
            ).sign(key, hashes.SHA256(), default_backend())
    except ValueError as e:
        LOG.critical('Error creating certificate: {}'.format(e.message))
        fname = options.output_file.name
        options.output_file.close()
        os.remove(fname)
        return 1
    # Write our certificate out to disk.
    options.output_file.write(cert.public_bytes(serialization.Encoding.DER))
    options.output_file.close()
    print('[\033[1;93mWARNING\033[1;0m]: Certificates generated with this tool are self-signed and for testing only',
            file=sys.stderr)
    if options.valid_time < 10 * 365.25:
        print('[\033[1;93mWARNING\033[1;0m]: This certificate is valid for {} days. For production,'
              'use certificates with at least 10 years validity.'.format(options.valid_time),
                file=sys.stderr)

    return 0

def add(options):
    if not hasattr(options, 'certificate'):
        LOG.critical('Cannot add certificate without certificate')
        return 1
    # Load the certificate
    cert = x509.load_der_x509_certificate(options.certificate.read(), default_backend())
    # Make sure the certificate uses SHA256.
    if not isinstance(cert.signature_hash_algorithm, hashes.SHA256):
        LOG.critical("In ({file}): Only SHA256 certificates are supported by the Mbed Cloud Update client at this time.".format(file=options.certificate.name))
        return 1
    if not isinstance(cert.public_key().curve, ec.SECP256R1):
        LOG.critical("In ({file}): Only secp256r1 (prime256v1) certificates are supported by the Mbed Cloud Update client at this time.".format(file=options.certificate.name))
        return 1

    fp = bytes(cert.fingerprint(hashes.SHA256()))
    newCertPath = os.path.join(defaults.certificatePath, binascii.b2a_hex(fp))
    with open(newCertPath, 'wb') as f:
        f.write(cert.public_bytes(serialization.Encoding.DER))
        LOG.info('Added certificate as: {}'.format(newCertPath))
    return 0

def read(options):
    pass

def main(options):
    return { "create": create,
      "add":    add,
    #   "read": parse
    }[options.cert_action](options)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
from __future__ import division
import sys, time, collections, hashlib, logging
from manifesttool import __version__ as uc_version
from manifesttool import codec
import os

from pyasn1.codec.der import encoder as der_encoder
from pyasn1.error import PyAsn1Error

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.kdf.hkdf import HKDF

from future.moves.urllib.request import urlopen, Request
from future.moves.urllib.parse import urlparse, urlencode
from future.moves.urllib.error import HTTPError, URLError

from builtins import bytes

LOG = logging.getLogger(__name__)

def fatal(message, *args):
    """Log a fatal error and exit with an unsuccessful code."""
    LOG.critical(message, *args)
    sys.exit(1)

def read_file(fname):
    with open(fname, 'rb') as f:
        return f.read()

def calculate_hash(options, m):
    md = hashes.Hash(hashes.SHA256(), backend=default_backend())
    md.update(m)
    return md.finalize()

def sha_hash(content):
    sha = hashlib.sha256()
    sha.update(content)
    return sha.digest()

def read_file_chunked(options, fd):
    data = fd.read(options.chunk_size)
    while data:
        yield data
        data = fd.read(options.chunk_size)

def hash_file(options, fname):
    if not hasattr(options,'hash') or not options.hash:
        options.hash = 'sha256'
    md = {
        'sha256' : hashes.Hash(hashes.SHA256(), backend=default_backend())
    }.get(options.hash, None)

    if not hasattr(options,'chunk_size') or not options.chunk_size:
        options.chunk_size = 4096
    with open(fname,'rb') as fd:
        for chunk in read_file_chunked(options, fd):
            md.update(chunk)
    return md.finalize()

def size_file(options, fname):
    return os.path.getsize(fname)

def getDevicePSK_HKDF(mode, masterKey, vendorId, classId, keyUsage):
    # Construct the device PSK
    # Device UUID is the IKM.
    hashAlg = {
        'aes-128-ctr-ecc-secp256r1-sha256' : hashes.SHA256,
        'none-ecc-secp256r1-sha256': hashes.SHA256,
        'none-none-sha256': hashes.SHA256,
        'none-psk-aes-128-ccm-sha256': hashes.SHA256,
        'psk-aes-128-ccm-sha256': hashes.SHA256
    }.get(mode)

    hkdf = HKDF(
        algorithm = hashAlg(),
        length = 128//8,
        # The master key is the salt
        salt = masterKey,
        # Device Vendor ID + Device Class ID is the info
        info = keyUsage + vendorId + classId,
        backend = default_backend()
    )
    return hkdf

def download_file(url):
    # Create request structure
    LOG.debug('Trying to download: {}'.format(url))
    try:
        req = Request(url)
    except ValueError as e:
        fatal('Client error. Could not download %r as the URL is invalid. Error: %r', url, str(e))
    req.add_header('User-Agent', 'ARM Update Client/%s)' % uc_version)

    # Read and return
    try:
        r = urlopen(req)
        content = r.read()
        r.close()
        return content
    except URLError as e:
        fatal('Could not download %r. Client error: %r', url, str(e))
    except HTTPError as e:
        fatal('Could not download %r. Server error: %r', url, str(e))

def todict(obj):
  """
  Recursively convert a Python object graph to sequences (lists)
  and mappings (dicts) of primitives (bool, int, float, string, ...)
  http://stackoverflow.com/a/22679824/503866
  """
  if sys.version_info.major == 3 and isinstance(obj, (str,bytes)):
    return obj
  elif sys.version_info.major == 2 and isinstance(obj, (str,bytes,unicode)):
    return obj
  elif isinstance(obj, dict):
    return dict((key, todict(val)) for key, val in obj.items() if val is not None)
  elif isinstance(obj, collections.Iterable):
    return [todict(val) for val in obj]
  elif hasattr(obj, '__dict__'):
    return todict(vars(obj))
  elif hasattr(obj, '__slots__'):
    return todict(dict((name, getattr(obj, name)) for name in getattr(obj, '__slots__')))
  return obj

def override_kwargs(obj, kwargs):
    for key, value in kwargs.items():
        if not value:
            continue
        if hasattr(obj, key):
            setattr(obj, key, value)

def encode(data, options, schema):
    encoder = {
        'der': der_encoder,
    }[options.encoding]

    try:
        asn = codec.obj2asn(data, schema)
        output = encoder.encode(asn)
        return output
    except PyAsn1Error as err:
        import traceback
        # exc_type, exc_value, exc_traceback = sys.exc_info()
        # traceback.print_stack()
        # traceback.print_tb(exc_traceback)
        LOG.critical('PyAsn1Error. This is propbably a bug. Could not encode '
                    'generated manifest.\nEnsure all fields are properly defined and '
                    'provided. Also check, in the case of the manifest changing, that '
                    'the generated "manifest_definition.py" file is up to date. The file can '
                    'be re-generated with the asn1py.py tool. Error message: "{0}"'.format(err))
        return None

def skipAhead(code):
    rc = 0
    if code == 0x81:
        rc = 1
    if code == 0x82:
        rc = 2
    if code == 0x83:
        rc = 3
    if code == 0x83:
        rc = 4
    return rc

def detect_version(input_file):
    # 32 bytes is currently sufficient to detect the manifest type.
    headerData = bytes(input_file.read(32))
    input_file.seek(0)
    # In both cases, the manifest starts with a DER SEQUENCE tag.
    if headerData[0] != 0x30:
        LOG.critical("input file is not a manifest.")
        return None
    # skip past the length
    pos = 2 + skipAhead(headerData[1])

    version = None
    # For version 1, the first object in the SEQUENCE should be another SEQUENCE
    if headerData[pos] == 0x30:
        version = '1'
    # For version 2+, a CMS wrapper is used, so the tag should be an OID tag
    if headerData[pos] == 0x06:
        version = '2'

    return version



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import json, logging
from builtins import bytes
from manifesttool.v1.parse import parse as parser_v1

LOG = logging.getLogger(__name__)

def skipAhead(code):
    rc = 0
    if code == 0x81:
        rc = 1
    if code == 0x82:
        rc = 2
    if code == 0x83:
        rc = 3
    if code == 0x83:
        rc = 4
    return rc

def parse(options):
    LOG.debug("Attempting to determine manifest version...")
    # 32 bytes is currently sufficient to detect the manifest type.
    headerData = bytes(options.input_file.read(32))
    options.input_file.seek(0)
    # In both cases, the manifest starts with a DER SEQUENCE tag.
    if headerData[0] != 0x30:
        LOG.critical("input file is not a manifest.")
        return None, None
    # skip past the length
    pos = 2 + skipAhead(headerData[1])

    version = None
    # For version 1, the first object in the SEQUENCE should be another SEQUENCE
    if headerData[pos] == 0x30:
        version = 1
    # For version 2+, a CMS wrapper is used, so the tag should be an OID tag
    if headerData[pos] == 0x06:
        version = 2

    if version == None:
        LOG.critical("No recognized manifest format found.")
        return None, None
    # For now, we will assume that 2+ means 2.
    parser = {
        1 : parser_v1
    }.get(version, None)
    if not parser:
        LOG.critical("Unrecognized manifest version.")
        return None, None
    return parser(options)


def main(options):
    decoded_manifest = parse(options)
    if not decoded_manifest:
        return 1

    indent = None if not options.pretty_json else 4

    # Write to output buffer/file
    options.output_file.write(str.encode(json.dumps(decoded_manifest, indent = indent)))

    # If we're writing to TTY, add a helpful newline
    if options.output_file.isatty():
        options.output_file.write(b'\n')



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import logging
import sys
LOG = logging.getLogger(__name__)

from manifesttool import prepare
from manifesttool import update_device as device

def main(options):
    if (hasattr(options,"psk") and options.psk) or (hasattr(options,"mac") and options.mac):
        LOG.critical('manifest-tool update commands are not currently enabled for PSK/MAC authentication.')
        return 1
    try:
        from mbed_cloud.update import UpdateAPI
        from mbed_cloud import __version__ as mc_version
        if mc_version < "1.2.6":
            LOG.warning("Your version of mbed-cloud-sdk may need updating: https://github.com/ARMmbed/mbed-cloud-sdk-python")
    except:
        LOG.critical('manifest-tool update commands require installation of the Mbed Cloud SDK: https://github.com/ARMmbed/mbed-cloud-sdk-python')
        return 1
    return {
        "prepare" : prepare.main,
        "device" : device.main
    }[options.update_action](options)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import sys, argparse, os
from manifesttool import defaults, __version__

class MainArgumentParser(object):

    def __init__(self):
        self.parser = self._make_parser()

    def _make_parser(self):
        parser = argparse.ArgumentParser(description = 'Create or transform a manifest.'
        ' Use {} [command] -h for help on each command.'.format(sys.argv[0]))

        # Add all top-level commands
        parser.add_argument('-l', '--log-level', choices=['debug','info','warning','exception'], default='info')
        parser.add_argument('--version', action='version', version=__version__,
            help='display the version'
        )
        subparsers = parser.add_subparsers(dest="action")
        subparsers.required = True
        create_parser = subparsers.add_parser('create', help='Create a new manifest')
        parse_parser = subparsers.add_parser('parse', help='Parse an existing manifest')
        verify_parser = subparsers.add_parser('verify', help='Verify an existing manifest')
        cert_parser = subparsers.add_parser('cert', help='Create or examine a certificate')
        init_parser = subparsers.add_parser('init', help='Set default values for manifests',
            description='''Configures default values for manifests.

            The manifest init command requires a certificate. If one is not provided, init will guide the user through
            the process of creating a certificate. Currently, the only supported type of created certificate is
            self-signed, which should only be used for testing/development purposes.

            When using an existing certificate, the user may supply a script to use in performing signing operations.
            This script is to support the use of exsiting Hardware Security Modules, etc.''')
        sign_parser = subparsers.add_parser('sign', help='Sign an existing manifest',
            description='''Signs an existing manifest with one additional signature. The manifest must first be
            validated, so defaults for validation are extracted from .manifest_tool.json when values are not supplied
            as arguments to the tool''')
        update_parser = subparsers.add_parser('update', help='Work with the Mbed Cloud Update service')

        # Options for creating a new manifest
        self._addCreateArgs(create_parser)

        # Options for parsing an existing manifest
        parse_parser.add_argument('-j','--pretty-json', action='store_true')
        parse_parser.add_argument('-e','--encoding', choices=['der'], default='der',
            help='Decode created manifest using the provided encoding scheme')
        self._add_input_file_arg(parse_parser)
        self._add_output_file_arg(parse_parser)

        # Options for verifying an existing manifest
        self._addVerifyArgs(verify_parser)

        cert_subparsers = cert_parser.add_subparsers(dest="cert_action")
        cert_subparsers.required = True

        cert_create_parser = cert_subparsers.add_parser('create',
            help='Create a certificate.\nWARNING: Certificates generated with this tool are self-signed and for testing only.',
            description='Create a certificate.\nWARNING: Certificates generated with this tool are self-signed and for testing only.'
            )
        # cert_read_parser = cert_subparsers.add_parser('read', help='Create a certificate')

        cert_create_parser.add_argument('-C','--country', help='Country of the subject', required=False)
        cert_create_parser.add_argument('-S','--state', help='State or province of the subject', required=False)
        cert_create_parser.add_argument('-L','--locality', help='Locality (city or region) of the subject', required=False)
        cert_create_parser.add_argument('-O','--organization', help='Subject organization that will hold the certificate', required=False)
        cert_create_parser.add_argument('-U','--common-name', help='Common name for the subject organization', required=False)
        cert_create_parser.add_argument('-V','--valid-time', help='time for the certificate to be valid, in days, starting now',
                                        type=int, required=True)
        cert_create_parser.add_argument('-K','--key-output', help='file to store the private key generated for the new certificate',
                                        type=argparse.FileType('wb'), required=True)
        cert_add_parser = cert_subparsers.add_parser('add',
            help='Add a certificate to the manifest-tool\'s list of certificates for use in manifest-tool verify.',
            description='Adds a certificate to the {} directory.'.format(defaults.certificatePath))

        cert_add_parser.add_argument('-c', '--certificate',
            help='DER-encoded certificate to add to the certificate query directory ({})'.format(defaults.certificatePath),
            type=argparse.FileType('rb'), required=True)

        self._add_output_file_arg(cert_create_parser)

        init_parser.add_argument('-c','--certificate', help='Provide an existing certificate to init',
                                type=argparse.FileType('rb'), required=False)
        init_existing_cert_signing = init_parser.add_mutually_exclusive_group(required=False)
        init_existing_cert_signing.add_argument('-k','--private-key',
                                help='Provide a private key file for the certificate provided with -c. ' +
                                     'This allows the manifest tool to perform manifest signing internally.',
                                type=argparse.FileType('rb'))
        # init_existing_cert_signing.add_argument('-s','--signing-script',
        #                         help='Provide a script that should be used for signing. ' +
        #                              'This allows signing with existing infrastructure. ' +
        #                              'The arguments to the script are: {fingerprint of the certificate} '+
        #                              '{hash of the manifest}.',type=argparse.FileType('rb'))

        init_vendor_group = init_parser.add_mutually_exclusive_group(required=True)
        init_vendor_group.add_argument('-V', '--vendor-id', help='')
        init_vendor_group.add_argument('-d', '--vendor-domain', help='')

        init_class_group = init_parser.add_mutually_exclusive_group(required=True)
        init_class_group.add_argument('-C', '--class-id', help='')
        init_class_group.add_argument('-m', '--model-name', help='')

        init_parser.add_argument('-S', '--server-address', help='Address of the API server for updates')
        init_parser.add_argument('-a', '--api-key', help='API Key for the API server')
        init_parser.add_argument('-q', '--quiet', help='Do not prompt for certificate fields', action='store_true')
        init_parser.add_argument('-f', '--force', action='store_true',
            help='Overwrite existing update_default_resources.c')
        init_parser.add_argument('--psk', action='store_true', help='initialize this project as a PSK authentication project')

        self._addVerifyArgs(sign_parser, ['input-file'])
        sign_source_group = sign_parser.add_mutually_exclusive_group(required=True)
        sign_source_group.add_argument('-m', '--manifest', type=argparse.FileType('rb'), default=sys.stdin)
        # sign_source_group.add_argument('-u', '--url')

        sign_parser.add_argument('-k', '--private-key', metavar='KEY',
            help='Supply a private key, or a shared secret for signing a created manifest',
            type=argparse.FileType('rb'))
        sign_parser.add_argument('-c','--certificate',
            help='Provide an existing certificate to reference in the manifest signature. This must match the private key',
            type=argparse.FileType('rb'))

        update_sub_parsers = update_parser.add_subparsers(dest='update_action')

        prepare_parser = update_sub_parsers.add_parser('prepare', help='Prepare an update',
            description='''Prepares an update for Mbed Cloud. This uploads the specified payload, creates a manifest
            and uploads the manifest.''')

        self._addCreateArgs(prepare_parser, ['output-file'])
        prepare_parser.add_argument('-o', '--output-file', metavar='FILE',
            help='Specify the output file for the manifest. A temporary file is used by default.',
            type=argparse.FileType('wb'), default=None)

        prepare_parser.add_argument('-n', '--payload-name',
            help='The reference name for the payload in Mbed Cloud. If no name is specified, then one will be created '
                 'using the file name of the payload and the current timestamp.')
        prepare_parser.add_argument('-d', '--payload-description',
            help='The description of the payload for use in Mbed Cloud.')
        prepare_parser.add_argument('--manifest-name',
            help='The reference name for the manifest in Mbed Cloud. If no name is specified, then one will be created '
                 'using the file name of the manifest and the current timestamp.')
        prepare_parser.add_argument('--manifest-description',
            help='The description of the manifest for use in Mbed Cloud.')
        prepare_parser.add_argument('-a', '--api-key', help='API Key for the Mbed Cloud')

        update_device_parser = update_sub_parsers.add_parser('device',
            help='Update a single device using its LwM2M Device ID',
            description='''Update a single device using its LwM2M Device ID. Note that settings may be provided in the
                manifest tool configuration file ({}). The API key MUST be provided in the Mbed Cloud Client
                configuration file.'''.format(defaults.config))
        self._addCreateArgs(update_device_parser, ['output-file'])

        update_device_parser.add_argument('-o', '--output-file', metavar='FILE',
            help='Specify the output file for the manifest. A temporary file is used by default.',
            type=argparse.FileType('wb'), default=None)

        update_device_parser.add_argument('-n', '--payload-name',
            help='The reference name for the payload in Mbed Cloud. If no name is specified, then one will be created '
                 'using the file name of the payload and the current timestamp.')
        update_device_parser.add_argument('-d', '--payload-description',
            help='The description of the payload for use in Mbed Cloud.')
        update_device_parser.add_argument('--manifest-name',
            help='The reference name for the manifest in Mbed Cloud. If no name is specified, then one will be created '
                 'using the file name of the manifest and the current timestamp.')
        update_device_parser.add_argument('--manifest-description',
            help='The description of the manifest for use in Mbed Cloud.')
        update_device_parser.add_argument('-D', '--device-id', help='The device ID of the device to update', required=True)
        update_device_parser.add_argument('--no-cleanup', action='store_true',
            help='''Don't delete the campaign, manifest, and firmware image from Mbed Cloud when done''')
        update_device_parser.add_argument('-T', '--timeout', type=int, default=-1,
            help='''Set the time delay before the manifest tool aborts the campaign. Use -1 for indefinite (this is the default).''')
        update_device_parser.add_argument('-a', '--api-key', help='API Key for the Mbed Cloud')

        return parser

    def _add_input_file_arg(self, parser):
        parser.add_argument('-i', '--input-file', metavar='FILE',
            help='Specify the input file. stdin by default',
            type=argparse.FileType('rb'), default=sys.stdin)

    def _add_output_file_arg(self, parser):
        output_default = sys.stdout
        if sys.version_info.major == 3:
            output_default = sys.stdout.buffer
        parser.add_argument('-o', '--output-file', metavar='FILE',
            help='Specify the output file. stdout by default',
            type=argparse.FileType('wb'), default=output_default)
    def _addCreateArgs(self, parser, exclusions=[]):
        parser.add_argument('-v', '--manifest-version', choices=['1'], default='1')
        if not 'private-key' in exclusions:
            parser.add_argument('-k', '--private-key', metavar='KEY',
                help='Supply a private key, or a shared secret for signing a created manifest',
                type=argparse.FileType('rb'))
        # create_parser.add_argument('-e','--encrypt-payload', choices=['aes-psk'])
        # create_parser.add_argument('-s','--payload-secret', metavar='SECRET',
        #     help='Secret data used for encryption, e.g. a pre-shared key or a private key.\
        #           The exact contents is dictated by the encryption method used. The secret may\
        #           be base64 encoded or may be a filename.')
        if not 'mac' in exclusions:
            parser.add_argument('--mac',
                help='Use Pre-Shared-Key MAC authentication, with a master key supplied in --private-key. '
                'A filter ID or Device Unique ID is also required to specify which devices should have TAGs created. '
                'These can be supplied in --filter-id or --device-urn.\n'
                'The manifest tool will create a PSK for each device based on the master key and the concatenation of three LwM2M values: \n'
                '    /10255/0/3 (Vendor ID)\n'
                '    /10255/0/4 (Device Class ID)\n'
                '    Device URN (Endpoint Client Name)',
                action='store_true')
            parser.add_argument('--filter-id', help='specify which devices to use.')
            parser.add_argument('--device-urn', action='append',
                help='Specify devices to target with the update by their URNs (Endpoint Client Name). '
                'The manifest tool will derive a PSK for the specified device based on the master key and the concatenation of: \n'
                '    Vendor ID\n'
                '    Device Class ID\n'
                '    Device URN\n'
                '--device-urn can be used multiple times to specify multiple devices.')
            parser.add_argument('--psk-table', help='Specify the file to use to store the PSK table. '
                'This file is used with the --mac argument in order to specify the output file for the pre-shared keys. '
                'The table is composed of three columns: device URN, WrappedManifestDigest, and WrappedPayloadKey.',
                type=argparse.FileType('wb'))
            parser.add_argument('--psk-table-encoding', help='', choices=['protobuf', 'text'], default='text')
        if not 'payload-key' in exclusions:
            parser.add_argument('--payload-key', help='supply the payload encryption key. This is the key that is used encrypt the payload. '
                'The payload key is encrypted for each device using a shared secret.')
        if not 'payload' in exclusions:
            parser.add_argument('-p', '--payload',
                help='Supply a local copy of the payload file.'
                     'This option overrides any payload file supplied in a `-i` argument.', metavar='FILE',
                type=argparse.FileType('rb'), required = True)
        if not 'uri' in exclusions:
            parser.add_argument('-u', '--uri',
                help='Supply the URI of the payload. '
                     'When a payload is uploaded to cloud storage, this is the URL of the payload. '
                     'This option overrides any URI supplied in a `-i` argument.')
            parser.add_argument('--url', help='Synonym for `--uri`', dest='uri')
        if not 'encoding' in exclusions:
            parser.add_argument('-c','--encoding', choices=['der'], default='der',
                help='Encode created manifest using the provided encoding scheme')
        if not 'hex' in exclusions:
            parser.add_argument('-x', '--hex', action='store_true',
                help='Output data in hex octets')
        if not 'c-file' in exclusions:
            parser.add_argument('-C', '--c-file', action='store_true',
                help='Output data as a C file')
        if not 'input-file' in exclusions:
            self._add_input_file_arg(parser)
        if not 'output-file' in exclusions:
            self._add_output_file_arg(parser)

    def _addVerifyArgs(self, verify_parser, exclusions=[]):
        if not 'pretty-json' in exclusions:
            verify_parser.add_argument('-j','--pretty-json', action='store_true')
        if not 'encoding' in exclusions:
            verify_parser.add_argument('-e','--encoding', choices=['der'], default='der',
                help='Decode created manifest using the provided encoding scheme')
        if not 'certificate-directory' in exclusions:
            verify_parser.add_argument('-d','--certificate-directory', default=defaults.certificatePath,
                help='A directory that contains the certificates necessary to validate a manifest. These should be named with their fingerprint.')
        if not 'vendor-id' in exclusions:
            verify_parser.add_argument('-V','--vendor-id', dest='vendorId', metavar='VENDORID',
                help='Set the vendor UUID that verify should expect.' )
        if not 'class-id' in exclusions:
            verify_parser.add_argument('-C','--class-id', dest='classId', metavar='CLASSID',
                help='Set the class UUID that verify should expect.' )
        if not 'input-file' in exclusions:
            self._add_input_file_arg(verify_parser)
        if not 'output-file' in exclusions:
            self._add_output_file_arg(verify_parser)

    def _verify_arguments(self):
        """Custom logic to ensure valid arguments are passed in"""
        # if self.options.action == "create":
        #     if self.options.encrypt_payload and not self.options.payload_secret:
        #         self.parser.error('A secret must be supplied with --payload-secret option when the --encrypt-payload option is in use.')
        pass

    def parse_args(self, args=None):
        self.options = self.parser.parse_args(args)
        self._verify_arguments()
        return self



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import logging, sys
LOG = logging.getLogger(__name__)

from manifesttool import create
from urllib3.exceptions import MaxRetryError
import copy
import time
import tempfile
import os
import os.path
import shutil


MAX_NAME_LEN = 128 # The update API has a maximum name length of 128, but this is not queriable.

tempdirname = None
manifest_file = None

def main_wrapped(options):
    try:
        from mbed_cloud.update import UpdateAPI
        import mbed_cloud.exceptions
    except:
        LOG.critical('manifest-tool update commands require installation of the Mbed Cloud SDK:'
                     ' https://github.com/ARMmbed/mbed-cloud-sdk-python')
        return 1
    LOG.debug('Preparing an update on Mbed Cloud')
    # upload a firmware
    api = None
    try:
        # If set use api key set in manifest-tool update.
        if hasattr(options, 'api_key') and options.api_key:
            tempKey = options.api_key
            config = {'api_key': tempKey}
            api = UpdateAPI(config)
        # Otherwise use API key set in manifest-tool init
        else: api = UpdateAPI()

    except ValueError:
        LOG.critical('API key is required to connect to the Update Service. It can be added using manifest-tool init -a'
                     ' <api key> or by manually editing .mbed_cloud_config.json')
        return 1
    if not options.payload_name:
        name = os.path.basename(options.payload.name) + time.strftime('-%Y-%m-%dT%H:%M:%S')
        LOG.info('Using {} as payload name.'.format(name))
        options.payload_name = name
    if len(options.payload_name) > MAX_NAME_LEN:
        LOG.critical(
            'Payload name is too long. Maximum length is {size}. ("{base}" <- {size}. overflow -> "{overflow}")'.format(
                size=MAX_NAME_LEN, base=options.payload_name[:MAX_NAME_LEN],
                overflow=options.payload_name[MAX_NAME_LEN:]))
        return 1

    if not options.manifest_name:
        name = os.path.basename(options.payload.name) + time.strftime('-%Y-%m-%dT%H:%M:%S-manifest')
        LOG.info('Using {} as manifest name.'.format(name))
        options.manifest_name = name

    if len(options.manifest_name) > MAX_NAME_LEN:
        LOG.critical(
            'Manifest name is too long. Maximum length is {size}. ("{base}" <- {size}. overflow -> "{overflow}")'.format(
                size=MAX_NAME_LEN, base=options.manifest_name[:MAX_NAME_LEN],
                overflow=options.manifest_name[MAX_NAME_LEN:]))
        return 1

    kwArgs = {}
    if options.payload_description:
        kwArgs['description'] = options.payload_description
    try:
        payload = api.add_firmware_image(
                        name = options.payload_name,
                    datafile = options.payload.name,
                    **kwArgs)
    except mbed_cloud.exceptions.CloudApiException as e:
        # TODO: Produce a better failuer message
        LOG.critical('Upload of payload failed with:\n{}'.format(e).rstrip())
        LOG.critical('Check API server URL "{}"'.format(api.config["host"]))
        return 1
    except MaxRetryError as e:
        LOG.critical('Upload of payload failed with:\n{}'.format(e))
        LOG.critical('Failed to establish connection to API-GW')
        LOG.critical('Check API server URL "{}"'.format(api.config["host"]))
        return 1

    LOG.info("Created new firmware at {}".format(payload.url))
    options.payload.seek(0)
    # create a manifest
    create_opts = copy.copy(options)
    create_opts.uri = payload.url
    create_opts.payload = options.payload
    if not (hasattr(create_opts, "output_file") and create_opts.output_file):
        global manifest_file
        manifest_file = open(os.path.join(tempdirname,'manifest'),'wb')
        create_opts.output_file = manifest_file

    rc = create.main(create_opts)
    create_opts.output_file.close()
    if rc:
        return rc

    kwArgs = {}
    if options.manifest_description:
        kwArgs['description'] = options.manifest_description

    # upload a manifest
    try:
        manifest = api.add_firmware_manifest(
                        name = options.manifest_name,
                    datafile = create_opts.output_file.name,
                    **kwArgs)
    except mbed_cloud.exceptions.CloudApiException as e:
        # TODO: Produce a better failuer message
        LOG.critical('Upload of manifest failed with:')
        print(e)
        return 1
    LOG.info('Created new manifest at {}'.format(manifest.url))
    LOG.info('Manifest ID: {}'.format(manifest.id))
    return 0

def main(options):
    global tempdirname
    tempdirname = tempfile.mkdtemp()
    RC = main_wrapped(options)
    if manifest_file:
        manifest_file.close()
    shutil.rmtree(tempdirname)
    return RC



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import json, logging, hashlib, time, sys, binascii, ecdsa, os, tempfile, uuid

from manifesttool.v1 import verify_signed_resource, verify_resource, verify_manifest_minimal, manifest_definition
from manifesttool import codec, utils, defaults

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import asymmetric
import cryptography

# All supported decoders
from pyasn1.codec.der import decoder as der_decoder

LOG = logging.getLogger(__name__)

commonParameters = [
    'timestamp'
]

def certificateQuery(options, fingerprints, URLs):
    '''Load a certificate from a directory that contains certificates named by fingerprint'''
    cert = None
    missingCertURLs = []
    # Load defaults
    defaultConfig = {}
    if os.path.exists(defaults.config):
        with open(defaults.config) as f:
            defaultConfig = json.load(f)

    for i,fingerprint in enumerate(fingerprints):
        path = os.path.join(options.certificate_directory, fingerprint)
        LOG.debug('Fetching certificate: {}'.format(path))
        #look up the fingerprint
        cert = None
        if os.path.exists(path) and os.path.isfile(path):
            cert = path
            break
        else:
            LOG.debug('Could not find {}, checking default certificates'.format(path))
            defaultCertificateList = defaultConfig.get('default-certificates', [])
            for defaultCertificate in defaultCertificateList:
                if 'file' in defaultCertificate:
                    LOG.debug('Checking {} ...'.format(defaultCertificate['file']))
                    with open(defaultCertificate['file'], 'rb') as certfile:
                        cert_data = certfile.read()
                        x509cert = x509.load_der_x509_certificate(cert_data, default_backend())
                        fp = x509cert.fingerprint(hashes.SHA256())
                        if binascii.b2a_hex(fp) == fingerprint:
                            cert = certfile.name
                            break
                        else:
                            LOG.debug('Fingerprint mismatch for {}\n'
                                         'Expected: {}\n'
                                         'Actual:   {}'.format(certfile.name, fingerprint, binascii.b2a_hex(fp)))
            if cert:
                break

                    # load the certificate and check its fingerprint.
        if not cert:
            LOG.debug('Failed to find certificate: {}'.format(path))
    else:
        return None
    # TODO: Fetch missing certificates
    # TODO: Validate certificate chain

    LOG.debug('Found certificate: {}'.format(cert))
    return cert

def verifyManifestV1(options, data, signedResource, resource, manifest):
    if manifest['manifestVersion'] != 'v1':
        LOG.critical('Not a version 1 manifest')
        return None
    LOG.info('Manifest version {}: OK'.format(manifest['manifestVersion']))
    # Extract the crypto mode
    if 'oid' in manifest['encryptionMode']:
        LOG.critical('Cannot verify Object ID encryptionMode')
        return None

    if not 'enum' in manifest['encryptionMode']:
        LOG.critical('No encryptionMode specified')
        return None

    emode = manifest['encryptionMode']['enum']
    modes = verify_manifest_minimal.Manifest().getComponentType()['encryptionMode'].getType().getComponentType()['enum'].getType().getNamedValues()
    if modes.getValue(emode) != None and emode != modes.getName(0):
        LOG.info('Encryption mode {}: OK'.format(emode))
    else:
        LOG.critical('Encryption mode {} not supported'.format(emode))
        return None
    # NOTE: Currently only modes 1-3 are supported and these modes only allow SHA256.
    # This means that the manifest is hashed with SHA256.

    # Load defaults
    defaultConfig = {}
    if os.path.exists(defaults.config):
        with open(defaults.config) as f:
            defaultConfig = json.load(f)

    # Verify UUIDs
    for i in ['vendorId', 'classId', 'deviceId']:
        if i in manifest:
            try:
                LOG.debug('Found {}: {}'.format(i,manifest[i]))
                manifest_id = uuid.UUID(manifest[i])
            except:
                #if len(uuid) != 0 and len(uuid) != 128/8:
                LOG.critical('UUIDs ({}) must be 0 or 128 bits'.format(i))
                return None
            expect_id = None
            if hasattr(options,i) and getattr(options,i):
                expect_id = uuid.UUID(getattr(options,i))
            else:
                expect_id = defaultConfig.get(i, None)
                if expect_id:
                    expect_id = uuid.UUID(expect_id)
                    LOG.debug('Using default {} from {}'.format(i,defaults.config))
            if expect_id:
                if manifest_id != expect_id:
                    LOG.critical('UUID mismatch for {}\n'
                                 'Expected: {}\n'
                                 'Actual:   {}'.format(i, expect_id, manifest_id))
                    return None
                LOG.info('UUID {}: OK'.format(i))
            else:
                LOG.warning('UUID {} not specified; ignoring'.format(i))

    # Verify the manifest hash
    if not 'signature' in signedResource:
        LOG.critical('Signature missing, but encryption mode {} requires a signature'.format(emode))
        return None
    c_hash = utils.sha_hash(signedResource['resource'])
    if not 'hash' in signedResource['signature']:
        LOG.critical('Manifest does not contain a hash')
        return None
    e_hash = binascii.a2b_hex(signedResource['signature']['hash'])
    if c_hash != e_hash:
        LOG.critical('Hash mismatch\nExpected: {}\nActual:   {}'.format(binascii.b2a_hex(e_hash),binascii.b2a_hex(c_hash)))
        return None
    LOG.debug('Manifest hash: {}'.format(binascii.b2a_hex(c_hash)))
    LOG.info('Maninfest hash: OK')

    if not options.certificateQuery:
        LOG.warning('No certificateQuery provided, will not verify signatures')
    if (emode == 'none-ecc-secp256r1-sha256' or emode == 'aes-128-ctr-ecc-secp256r1-sha256') and options.certificateQuery:
        if not 'signatures' in signedResource['signature']:
            LOG.critical('Signature missing, but encryption mode {} requires a signature'.format(emode))
            return None
        for signature in signedResource['signature']['signatures']:
            if not 'certificates' in signature:
                LOG.critical('A certificate reference is mandatory in a signature')
                return None
            if len(signature['certificates']) == 0:
                LOG.critical('At least one certificate reference is mandatory in a signature')
                return None
            if not 'fingerprint' in signature['certificates'][0] or len(signature['certificates'][0]['fingerprint']) == 0:
                LOG.critical('A fingerprint is mandatory in a certificate')
                return None
            LOG.debug('Verifying signature by {}'.format(signature['certificates'][0]['fingerprint']))
            fingerprints = [x['fingerprint'] for x in signature['certificates']]
            URLs = [x['uri'] if 'uri' in x else '' for x in signature['certificates']]
            certfile = options.certificateQuery(options, fingerprints, URLs)
            if certfile == None: # TODO: Option for mandatory certificate verification
                if options.mandatory_signature:
                    LOG.critical('Could not find certificate chain matching {}'.format(fingerprints[0]))
                    return None
                LOG.warning('Could not find certificate chain matching {}'.format(fingerprints[0]))
            else:
                #if not options.ecdsaVerify(cert, signature):
                # ok = False
                # vk = ecdsa.VerifyingKey.from_pem(keypem)
                # ok = vk.verify(binascii.a2b_hex(signature['signature']),
                #           signedResource['resource'],
                #           hashfunc=hashlib.sha256,
                #           sigdecode=ecdsa.util.sigdecode_der)
                LOG.debug('Opening {} ...'.format(certfile))
                with open(certfile,'rb') as cert:
                    ok = options.ecdsaVerify(cert,signature['signature'],e_hash)
                    if not ok:
                        LOG.critical('Signature verification failed')
                        return None
                    LOG.info('Signature by {}: OK'.format(fingerprints[0]))


    # Decode the full manifest.
    full_manifest = {
        "der": lambda d: codec.bin2obj(data, manifest_definition.SignedResource(), der_decoder),
    }[options.encoding](data)
    LOG.debug('Parsed whole manifest from {}-encoded binary object'.format(options.encoding))

    full_manifest = full_manifest['resource']['resource']['manifest']
    # TODO: Measure nonce entropy
    if not 'nonce' in full_manifest:
        LOG.critical('A nonce is required in the manifest')
        return None
    else:
        nonce = binascii.a2b_hex(full_manifest['nonce'])
        if len(nonce)*8 != 128:
            LOG.critical('Nonce must be 128 bits. Got {}: {}'.format(len(nonce)*8, binascii.b2a_hex(nonce)))
            return None
    LOG.info('nonce: {} OK'.format(full_manifest['nonce']))

    # Call vendor-supplied Vendor Info Validator
    if hasattr(options, 'validateVendorInfo'):
        if options.validateVendorInfo(options, full_manifest['vendorInfo']):
            LOG.critical('Vendor Info Validation failed')
            return None
        LOG.info('VendorInfo: OK')

    # Extract apply period
    applyPeriod = full_manifest.get('applyPeriod')

    # must have either a payload or a dependency
    if not 'payload' in full_manifest and (
            not 'dependencies' in full_manifest or len(full_manifest['dependencies']) == 0):
        LOG.critical('Manifest must contain either a dependency or a payload')
        sys.exit(1)

    # Verify the payload
    if 'payload' in full_manifest:
        payload = full_manifest['payload']
        LOG.debug('Manifest contains a payload')
        # Verify the payload format
        if 'enum' in payload['format']:
            enum = payload['format']['enum']
            payloadFormats = manifest_definition.PayloadDescription().getComponentType()['format'].getType().getComponentType()['enum'].getType().getNamedValues()
            if payloadFormats.getValue(enum) == None or enum == payloadFormats.getName(0):
                LOG.critical('Payload format not recognized')
                return None
            LOG.info('Payload format {}: OK'.format(enum))
        elif 'objectId' in payload['format']:
            LOG.warning('Cannot verify Object ID payload format')
        else:
            LOG.critical('Payload does not contain a format')
            return None

        if emode == 'aes-128-ctr-ecc-secp256r1-sha256':
            if not 'encryptionInfo' in payload:
                LOG.critical('Encryption info must be present for encrypted payload distribution')
                return None
            cryptinfo = payload['encryptionInfo']
            # Validate the encryption information
            # Validate the init vector:
            if not 'initVector' in cryptinfo or len(binascii.a2b_hex(cryptinfo['initVector'])) != 128/8:
                LOG.critical('When using aes-128-ctr-ecc-secp256r1-sha256, a 128-bit AES initialization vector is mandatory')
                if 'initVector' in cryptinfo:
                    iv = binascii.a2b_hex(cryptinfo['initVector'])
                    LOG.critical('Expected 128 bits, got {}: {}'.format(len(iv)*8, binascii.b2a_hex(iv)))
                return None
            # TODO: Verify the entropy of the init vector

            # Determine the key mode
            kmode = 0
            # Options:
            if 'key' in cryptinfo['id'] and 'cipherKey' in cryptinfo['key']:
                if len(cryptinfo['key']['cipherKey']) == 0:
                    # Select a preshared local key
                    kmode = 1
                    LOG.debug('Using local preshared key for decryption')
                else:
                    # Select a preshared local key & decrypt a session key
                    kmode = 2
                    LOG.debug('Using local preshared key to decrypt the payload key')
            # Select certificate & decrypt a session key (Single-device only)
            elif 'certificate' in cryptinfo['id'] and 'cipherKey' in cryptinfo['key']:
                kmode = 3
                LOG.debug('Using ECDH to decrypt the device key')
            # Select a certificate & a keytable
            elif 'certificate' in cryptinfo['id'] and 'keyTable' in cryptinfo['key']:
                kmode = 4
                LOG.debug('Using ECDH to decrypt the payload key from the key table')

            if kmode == 0:
                LOG.critical('Unrecognized key distribution mode')
                return None

            # NOTE: It is not possible to verify payload without a device key when it is encrypted.
        # Verify the storage identifier
        if not 'storageIdentifier' in payload or len(payload['storageIdentifier']) == 0:
            LOG.critical('storageIdentifier must be provided')
            return None
        # Verify the resource reference
        if not 'reference' in payload:
            LOG.critical('A resource reference is mandatory in a payload-bearing manifest')
            return None
        if not 'hash' in payload['reference'] or len(payload['reference']['hash']) == 0:
            LOG.critical('A resource hash is mandatory in a payload reference')
            return None
        if not 'size' in payload['reference'] or payload['reference']['size'] == 0:
            LOG.critical('Zero-size resources are not permitted')
            return None

        if 'uri' in payload['reference']:
             LOG.debug('Payload refers to URI: {}'.format(payload['reference']['uri']))
             # TODO: verify payload URI

        # Do not verify the version string; it is for presentation only.

    # TODO: Validate aliases
    # TODO: Validate dependencies

    return {'timestamp':full_manifest['timestamp'], 'applyPeriod': applyPeriod}


def verifyManifest(options):
    data = options.input_file.read()
    LOG.debug('Read {} bytes of encoded data. Will try to decode...'.format(len(data)))

    # TODO: Verify the DER structure's encoding

    # Extract the signed resource.
    signed_resource_data = {
        "der": lambda d: codec.bin2obj(d, verify_signed_resource.SignedResource(), der_decoder),
    }[options.encoding](data)
    LOG.debug('Decoded SignedResource from {} encoded binary'.format(options.encoding))

    # Verify that the content *is* a manifest.
    # NOTE: This requires a parsing pass of the resource.
    resource_data = {
        "der": lambda d: codec.bin2obj(d, verify_resource.Resource(), der_decoder),
    }[options.encoding](signed_resource_data['resource'])
    LOG.debug('Decoded Resource from {} encoded binary'.format(options.encoding))

    if resource_data['resourceType'] != 'manifest':
        LOG.critical('The supplied file does not contain a manifest')
        return 1

    # Extract some relevant manifest information:
    manifest_data = {
        "der": lambda d: codec.bin2obj(d, verify_manifest_minimal.Manifest(), der_decoder),
    }[options.encoding](resource_data['resource'])
    LOG.debug('Decoded Manifest from {} encoded binary'.format(options.encoding))

    # Select a manifest format decoder based on the manifest version.
    def noVersionError(options, data, signedResource, resource, manifest):
        LOG.critical('Unsupported manifest version: {}'.format(manifest_data['manifestVersion']))
        None

    manifest_verification_data = {
        'v1': verifyManifestV1
    }.get(manifest_data['manifestVersion'],noVersionError)(options, data, signed_resource_data, resource_data, manifest_data)

    if manifest_verification_data == None:
        return 1

    # Verify the manifest timestamp is sane

    # The manifest timestamp should not be in the future, nor before the first release of this tool.
    # manifest_timestamp = decoded_data['resource']['resource']['manifest']['timestamp']
    manifest_timestamp = manifest_verification_data['timestamp']
    systime = int(time.time())
    if manifest_timestamp > systime:
        LOG.critical('Manifests MUST not be timestamped in the future.\n'
                     'Expected timestamp < {} ({})\n'
                     'Actual timestamp:    {} ({})'.format(
                     systime, time.ctime(systime), manifest_timestamp, time.ctime(manifest_timestamp)))

    release_time = time.mktime(time.strptime('2016-11-22T00:00:00',"%Y-%m-%dT%H:%M:%S"))
    if manifest_timestamp < int(release_time):
        LOG.critical('Manifests MUST not be timestamped before 2017.\n'
                     'Expected timestamp > {} ({})\n'
                     'Actual timestamp:    {} ({})'.format(
                     int(release_time), time.ctime(int(release_time)), manifest_timestamp, time.ctime(manifest_timestamp)))
    LOG.info('Timestamp {} < {} < {}: OK'.format(int(release_time), manifest_timestamp, systime))

    # Verify applyPeriod
    if 'applyPeriod' in manifest_verification_data and manifest_verification_data['applyPeriod']:
        applyPeriod = manifest_verification_data['applyPeriod']
        if release_time > applyPeriod['validFrom'] or \
                applyPeriod['validFrom'] > applyPeriod['validTo'] or \
                applyPeriod['validTo'] > systime:
            LOG.critical('Apply perion outside expected bounds: Expected:\n{} <= {} <= {} <= {}'.format(
                release_time, applyPeriod['validFrom'], applyPeriod['validTo'], systime
            ))
            return 1
        LOG.info('Apply period {} to {} OK'.format(applyPeriod['validFrom'], applyPeriod['validTo']))


    # Verify each URL

    # Fetch each file from each listed URL (may not be possible, depending on network architecture)
    # Validate the hash of each file from each listed URL

    # Write to output buffer/file
    # indent = None if not options.pretty_json else 2
    # options.output_file.write(str.encode(json.dumps(decoded_data, indent = indent)))

    # If we're writing to TTY, add a helpful newline
    # if options.output_file.isatty():
    #     options.output_file.write(b'\n')

    return 0

def cryptographyEcdsaVerify(certfile, sig, sha):
    LOG.debug('Reading {}'.format(certfile.name))
    cert_data = certfile.read()
    cert = x509.load_der_x509_certificate(cert_data, default_backend())
    pubkey = cert.public_key()
    bsig = binascii.a2b_hex(sig)
    LOG.debug('Verifying...\n'
              'Signature: ({sigtype}){sig}\n'
              'Hash:      ({shatype}){sha}'.format(
              sigtype=type(bsig),sig=repr(bsig),shatype=type(sha),sha=repr(sha)))
    try:
        pubkey.verify(
            bsig,
            sha,
            asymmetric.ec.ECDSA(asymmetric.utils.Prehashed(hashes.SHA256()))
        )
    except cryptography.exceptions.InvalidSignature as e:
        return False
    return True


def main(options):

    options.ecdsaVerify = cryptographyEcdsaVerify

    if not options.certificate_directory:
        options.certificate_directory = defaults.certificatePath
    options.certificateQuery = certificateQuery
    options.mandatory_signature = True
    # if not hasattr(options, 'verifyScript'):

    return verifyManifest(options)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from setuptools import setup, find_packages
import pip
import manifesttool
import os


try: # for pip >= 10
    from pip._internal.req import parse_requirements
    from pip._internal.download import PipSession
except ImportError: # for pip <= 9.0.3
    from pip.req import parse_requirements
    from pip.download import PipSession

install_reqs = parse_requirements('requirements.txt', session=PipSession())
reqs = [str(r.req) for r in install_reqs]

if os.name == 'nt':
    entry_points={
        "console_scripts": [
            "manifest-tool=manifesttool.clidriver:main",
        ],
    }
    scripts = []
else:
    platform_deps = []
    # entry points are nice, but add ~100ms to startup time with all the
    # pkg_resources infrastructure, so we use scripts= instead on unix-y
    # platforms:
    scripts = ['bin/manifest-tool', ]
    entry_points = {}

setup(
    name='manifest-tool',
    version=manifesttool.__version__,
    description='Tool/lib to create and parse manifests',
    long_description=open("README.md").read(),
    url='https://github.com/ARMmbed/update-client-manifest-manager/manifestTool',
    author='Brendan Moran',
    author_email='brendan.moran@arm.com',
    packages=find_packages(exclude=['tests*']),
    zip_safe=False,
    scripts=scripts,
    entry_points=entry_points,
    install_requires=reqs
)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import logging, sys

from manifesttool.argparser import MainArgumentParser
from manifesttool import create, parse, verify, cert, init, sign
from manifesttool import update
import colorama
colorama.init()


LOG = logging.getLogger(__name__)
LOG_FORMAT='[%(levelname)s] %(asctime)s - %(name)s - %(message)s'

def main():
    driver = CLIDriver()
    return driver.main()

class CLIDriver(object):

    def __init__(self):
        self.options = MainArgumentParser().parse_args().options

        log_level = {
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warning': logging.WARNING,
            'exception': logging.CRITICAL
        }[self.options.log_level]
        logging.basicConfig(level=log_level,
                            format=LOG_FORMAT,
                            datefmt='%Y-%m-%d %H:%M:%S')
        logging.addLevelName( logging.INFO, "\033[1;32m%s\033[1;0m" % logging.getLevelName(logging.INFO))
        logging.addLevelName( logging.WARNING, "\033[1;93m%s\033[1;0m" % logging.getLevelName(logging.WARNING))
        logging.addLevelName( logging.CRITICAL, "\033[1;31m%s\033[1;0m" % logging.getLevelName(logging.CRITICAL))

        LOG.debug('CLIDriver created. Arguments parsed and logging setup.')

    def main(self):
        rc = { "create": create.main,
          "parse": parse.main,
          "verify": verify.main,
          "cert": cert.main,
          "init": init.main,
          "update" : update.main,
          "sign": sign.main
        }[self.options.action](self.options) or 0

        sys.exit(rc)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
#
# This file has been generated using asn1ate (v <unknown>) from './ASN.1/v1/manifest-1.0.0'
# Last Modified on 2017-11-21 17:47:43
from pyasn1.type import univ, char, namedtype, namedval, tag, constraint, useful


class Uri(char.UTF8String):
    pass


class Bytes(univ.OctetString):
    pass


class UUID(univ.OctetString):
    pass


class Payload(univ.OctetString):
    pass


class CertificateReference(univ.Sequence):
    pass


CertificateReference.componentType = namedtype.NamedTypes(
    namedtype.NamedType('fingerprint', Bytes()),
    namedtype.NamedType('uri', Uri())
)


class SignatureBlock(univ.Sequence):
    pass


SignatureBlock.componentType = namedtype.NamedTypes(
    namedtype.NamedType('signature', univ.OctetString()),
    namedtype.NamedType('certificates', univ.SequenceOf(componentType=CertificateReference()))
)


class MacBlock(univ.Sequence):
    pass


MacBlock.componentType = namedtype.NamedTypes(
    namedtype.NamedType('pskID', univ.OctetString()),
    namedtype.NamedType('keyTableVersion', univ.Integer()),
    namedtype.OptionalNamedType('keyTableIV', univ.OctetString()),
    namedtype.OptionalNamedType('keyTableRef', char.UTF8String()),
    namedtype.NamedType('keyTableIndexSize', univ.Integer()),
    namedtype.NamedType('keyTableRecordSize', univ.Integer())
)


class KeyTableEntry(univ.Sequence):
    pass


KeyTableEntry.componentType = namedtype.NamedTypes(
    namedtype.OptionalNamedType('hash', univ.OctetString().subtype(implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0))),
    namedtype.OptionalNamedType('payloadKey', univ.OctetString().subtype(implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1)))
)


class ResourceSignature(univ.Sequence):
    pass


ResourceSignature.componentType = namedtype.NamedTypes(
    namedtype.NamedType('hash', univ.OctetString()),
    namedtype.NamedType('signatures', univ.SequenceOf(componentType=SignatureBlock())),
    namedtype.OptionalNamedType('macs', univ.SequenceOf(componentType=MacBlock()))
)


class ResourceReference(univ.Sequence):
    pass


ResourceReference.componentType = namedtype.NamedTypes(
    namedtype.NamedType('hash', univ.OctetString()),
    namedtype.OptionalNamedType('uri', Uri()),
    namedtype.NamedType('size', univ.Integer())
)


class ResourceAlias(univ.Sequence):
    pass


ResourceAlias.componentType = namedtype.NamedTypes(
    namedtype.NamedType('hash', univ.OctetString()),
    namedtype.NamedType('uri', Uri())
)


class PayloadDescription(univ.Sequence):
    pass


PayloadDescription.componentType = namedtype.NamedTypes(
    namedtype.NamedType('format', univ.Choice(componentType=namedtype.NamedTypes(
        namedtype.NamedType('enum', univ.Enumerated(namedValues=namedval.NamedValues(('undefined', 0), ('raw-binary', 1), ('cbor', 2), ('hex-location-length-data', 3), ('elf', 4)))),
        namedtype.NamedType('objectId', univ.ObjectIdentifier())
    ))
    ),
    namedtype.OptionalNamedType('encryptionInfo', univ.Sequence(componentType=namedtype.NamedTypes(
        namedtype.NamedType('initVector', univ.OctetString()),
        namedtype.NamedType('id', univ.Choice(componentType=namedtype.NamedTypes(
            namedtype.NamedType('key', univ.OctetString()),
            namedtype.NamedType('certificate', CertificateReference())
        ))
        ),
        namedtype.OptionalNamedType('key', univ.Choice(componentType=namedtype.NamedTypes(
            namedtype.NamedType('keyTable', Uri()),
            namedtype.NamedType('cipherKey', univ.OctetString())
        ))
        )
    ))
    ),
    namedtype.NamedType('storageIdentifier', char.UTF8String()),
    namedtype.NamedType('reference', ResourceReference()),
    namedtype.OptionalNamedType('version', char.UTF8String())
)


class Manifest(univ.Sequence):
    pass


Manifest.componentType = namedtype.NamedTypes(
    namedtype.NamedType('manifestVersion', univ.Enumerated(namedValues=namedval.NamedValues(('v1', 1)))),
    namedtype.OptionalNamedType('description', char.UTF8String()),
    namedtype.NamedType('timestamp', univ.Integer()),
    namedtype.NamedType('vendorId', UUID()),
    namedtype.NamedType('classId', UUID()),
    namedtype.NamedType('deviceId', UUID()),
    namedtype.NamedType('nonce', univ.OctetString()),
    namedtype.NamedType('vendorInfo', univ.OctetString()),
    namedtype.OptionalNamedType('applyPeriod', univ.Sequence(componentType=namedtype.NamedTypes(
        namedtype.NamedType('validFrom', univ.Integer()),
        namedtype.NamedType('validTo', univ.Integer())
    ))
    ),
    namedtype.NamedType('applyImmediately', univ.Boolean()),
    namedtype.NamedType('encryptionMode', univ.Choice(componentType=namedtype.NamedTypes(
        namedtype.NamedType('enum', univ.Enumerated(namedValues=namedval.NamedValues(('invalid', 0), ('aes-128-ctr-ecc-secp256r1-sha256', 1), ('none-ecc-secp256r1-sha256', 2), ('none-none-sha256', 3), ('none-psk-aes-128-ccm-sha256', 4), ('aes-128-ccm-psk-sha256', 5)))),
        namedtype.NamedType('objectId', univ.ObjectIdentifier())
    ))
    ),
    namedtype.NamedType('aliases', univ.SequenceOf(componentType=ResourceAlias())),
    namedtype.NamedType('dependencies', univ.SequenceOf(componentType=ResourceReference())),
    namedtype.OptionalNamedType('payload', PayloadDescription())
)


class Resource(univ.Sequence):
    pass


Resource.componentType = namedtype.NamedTypes(
    namedtype.OptionalNamedType('uri', Uri()),
    namedtype.NamedType('resourceType', univ.Enumerated(namedValues=namedval.NamedValues(('manifest', 0), ('payload', 1)))),
    namedtype.NamedType('resource', univ.Choice(componentType=namedtype.NamedTypes(
        namedtype.NamedType('manifest', Manifest()),
        namedtype.NamedType('payload', Payload())
    ))
    )
)


class SignedResource(univ.Sequence):
    pass


SignedResource.componentType = namedtype.NamedTypes(
    namedtype.NamedType('resource', Resource()),
    namedtype.NamedType('signature', ResourceSignature())
)





# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright 2016-2017 ARM Limited or its affiliates
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
#
# This file has been generated using asn1ate (v <unknown>) from './manifesttool/verify/manifest.asn'
# Last Modified on 2017-01-10 11:40:03
from pyasn1.type import univ, char, namedtype, namedval, tag, constraint, useful


class UUID(univ.OctetString):
    pass


class Manifest(univ.Sequence):
    pass


Manifest.componentType = namedtype.NamedTypes(
    namedtype.NamedType('manifestVersion', univ.Enumerated(namedValues=namedval.NamedValues(('v1', 1)))),
    namedtype.OptionalNamedType('description', char.UTF8String()),
    namedtype.NamedType('timestamp', univ.Integer()),
    namedtype.NamedType('vendorId', UUID()),
    namedtype.NamedType('classId', UUID()),
    namedtype.NamedType('deviceId', UUID()),
    namedtype.NamedType('nonce', univ.OctetString()),
    namedtype.NamedType('vendorInfo', univ.OctetString()),
    namedtype.OptionalNamedType('applyPeriod', univ.Sequence(componentType=namedtype.NamedTypes(
        namedtype.NamedType('validFrom', univ.Integer()),
        namedtype.NamedType('validTo', univ.Integer())
    ))
    ),
    namedtype.NamedType('applyImmediately', univ.Boolean()),
    namedtype.NamedType('encryptionMode', univ.Choice(componentType=namedtype.NamedTypes(
        namedtype.NamedType('enum', univ.Enumerated(namedValues=namedval.NamedValues(('invalid', 0), ('aes-128-ctr-ecc-secp256r1-sha256', 1), ('none-ecc-secp256r1-sha256', 2), ('none-none-sha256', 3), ('none-psk-aes-128-ccm-sha256', 4), ('aes-128-ccm-psk-sha256', 5)))),
        namedtype.NamedType('objectId', univ.ObjectIdentifier())
    ))
    ),
    namedtype.NamedType('aliases', univ.Any()),
    namedtype.NamedType('dependencies', univ.Any()),
    namedtype.OptionalNamedType('payload', univ.Any())
)
